"""Functions for indexing named objects in JSON to be imported."""

from __future__ import annotations

from collections import OrderedDict
from copy import deepcopy
from re import findall
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from nettlesome.predicates import StatementTemplate
from authorityspoke.io import text_expansion


RawPredicate = Dict[str, Union[str, bool]]
RawFactor = Dict[str, Union[RawPredicate, Sequence[Any], str, bool]]


class Mentioned(OrderedDict):
    """Index of cross-referenced objects, keyed to phrases that reference them."""

    def insert_by_name(self, obj: Dict) -> None:
        """Add record to dict, using value of record's "name" field as the dict key."""
        self[obj["name"]] = obj.copy()
        self[obj["name"]].pop("name")
        return None

    def get_by_name(self, name: str) -> Dict:
        """
        Convert retrieved record so name is a field rather than the key for the whole record.

        :param name:
            the name of the key where the record can be found in the Mentioned dict.

        :returns:
            the value stored at the key "name", plus a name field.
        """
        if not self.get(name):
            raise ValueError(
                f'Name "{name}" not found in the index of mentioned Factors'
            )
        value = {"name": name}
        value.update(self[name])
        return value

    def sorted_by_length(self) -> Mentioned:
        """
        Sort dict items from longest to shortest.

        Used to ensure that keys nearer the start can't be substrings of later keys.
        """
        return Mentioned(sorted(self.items(), key=lambda t: len(t[0]), reverse=True))

    def __str__(self):
        return f"Mentioned({str(dict(self))})"

    def __repr__(self):
        return f"Mentioned({repr(dict(self))})"


def assign_name_from_content(obj: Dict) -> str:
    r"""
    Use the content to assign a name to any Fact that lacks one.

    This can fail if the parser replaces placeholder text with another placeholder

    :param obj:
        object loaded from JSON to make a :class:`.Factor` or :class:`.Holding`

    :returns:
        a new name
    """

    if obj.get("terms"):
        template = StatementTemplate(obj["predicate"]["content"], make_singular=False)
        placeholders = template.placeholders
        if any(placeholders):
            substitutions = dict(zip(placeholders, obj["terms"]))
            content_for_name = template.substitute(substitutions)
        else:
            content_for_name = obj["predicate"]["content"]
            for context_factor in obj["terms"]:
                content_for_name = content_for_name.replace("{}", context_factor, 1)
    else:
        content_for_name = obj["predicate"]["content"]
    false_modifier = "false " if obj["predicate"].get("truth") is False else ""
    return f"{false_modifier}{content_for_name}".replace("{", "").replace("}", "")


def assign_name_for_evidence(obj: Dict) -> str:
    r"""
    Return an appropriate name for Evidence.

    :param obj: an unloaded :class:`.Evidence` object

    :returns: a name for the Evidence
    """
    name = "evidence"
    if obj.get("exhibit"):
        name += f' of {obj["exhibit"]}'
    if obj.get("to_effect"):
        name += f' to the effect that {obj["to_effect"]}'
    return name


def assign_name_for_pleading(obj: Dict) -> str:
    r"""
    Return an appropriate name for a Pleading.

    :param obj: an unloaded :class:`.Pleading`

    :returns: a name for the Pleading
    """
    name = "pleading"
    if obj.get("filer"):
        name += f' filed by {obj["filer"]}'
    return name


def create_name_for_factor(obj: Dict) -> str:
    """
    Determine what kind of RawFactor the input is and return an appropriate name.

    :param obj: an unloaded :class:`.Factor`

    :returns: a name for the Factor
    """

    if (
        obj.get("content")  # Predicates don't need name
        or obj.get("rule")
        or obj.get("procedure")
        or obj.get("outputs")  # Procedures, Rules, and Holdings don't need names
        or obj.get("exact")
        or obj.get("prefix")
        or obj.get("suffix")  # Text Selectors don't need names
    ):
        return ""
    elif obj.get("predicate", {}).get("content"):
        name = assign_name_from_content(obj)
    elif obj.get("exhibit") or (obj.get("type") and obj["type"].lower()) == "evidence":
        name = assign_name_for_evidence(obj)
    elif obj.get("type") and obj["type"].lower() == "pleading":
        name = assign_name_for_pleading(obj)
    else:
        raise NotImplementedError
    if obj.get("absent") is True:
        name = f"absent {name}"
    return name


def ensure_factor_has_name(obj: Dict) -> Dict:
    """
    Add a name to a RawFactor if it doesn't already have one.

    :param obj: an unloaded :class:`.Factor`

    :returns: the same :class:`.Factor` with a name field added
    """
    if not obj.get("name"):
        new_name = create_name_for_factor(obj)
        if new_name:
            obj["name"] = new_name
    return obj


def update_name_index_from_terms(terms: List[RawFactor], mentioned: Mentioned):
    r"""
    Update name index from a list of :class:`.RawFactor`\s or strings referencing them.

    :param terms:
        a list of :class:`.RawFactor`\s or strings referencing them. Both
        :class:`.RawFactor`\s and strings may exist in the list.

    :param mentioned:
        :class:`.RawFactor`\s indexed by name for retrieval when loading objects
        using a Marshmallow schema.

    :returns:
        an updated "mentioned" name index
    """

    for factor in terms:
        if isinstance(factor, str):
            factor_name = factor
            if factor_name not in mentioned:
                raise ValueError(
                    f"Unable to expand referenced Factor '{factor_name}'. "
                    "It doesn't exist in the index of mentioned Factors."
                )
        else:
            factor_name = factor.get("name")
            if factor_name and factor_name not in mentioned:
                mentioned.insert_by_name(factor)
    return mentioned.sorted_by_length()


RawContextFactors = List[Union[RawFactor, str]]


def update_name_index_from_bracketed_phrases(
    content: str, mentioned: Mentioned
) -> Mentioned:
    pattern = r"(?<!\$)\{([^\{]+)\}"  # matches bracketed text not preceded by $
    entities_as_text = findall(pattern, content)
    entities_as_text.sort(key=len, reverse=True)

    for entity_name in entities_as_text:
        if entity_name not in mentioned:
            entity = {"type": "Entity", "name": entity_name}
            mentioned.insert_by_name(entity)
    return mentioned


def update_name_index_from_fact_content(
    obj: RawFactor, mentioned: Mentioned
) -> Tuple[RawFactor, Mentioned]:
    r"""
    Update index of mentioned Factors from Factors mentioned in Fact's content, and vice versa.

    :param obj:
        data from JSON to be loaded as a :class:`.Factor`
    :param mentioned:
        :class:`.RawFactor`\s indexed by name for retrieval when loading objects
        using a Marshmallow schema.

    :returns:
        both 'obj' and 'mentioned', updated with values from each other
    """

    predicate: RawPredicate = obj.get("predicate", {})
    content: str = predicate.get("content", "")
    if content:
        terms: RawContextFactors = obj.get("terms", [])
        mentioned = update_name_index_from_bracketed_phrases(
            content=content, mentioned=mentioned
        )
        mentioned = update_name_index_from_terms(terms, mentioned)

        for name in mentioned.keys():
            if name in content and name != content:
                (content, terms,) = text_expansion.add_found_context(
                    content=content,
                    terms=terms,
                    factor=mentioned.get_by_name(name),
                )
        obj["terms"] = terms
        obj["predicate"]["content"] = content
    return obj, mentioned


def update_name_index_with_factor(
    obj: RawFactor, mentioned: Mentioned
) -> Tuple[Union[str, RawFactor], Mentioned]:
    r"""
    Update index of mentioned Factors with 'obj', if obj is named.

    If there is already an entry in the mentioned index with the same name
    as obj, the old entry won't be replaced. But if any additional text
    anchors are present in the new obj, the anchors will be added.

    If obj has a name, it will be collapsed to a name reference.

    :param obj:
        data from JSON to be loaded as a :class:`.Factor`

    :param mentioned:
        :class:`.RawFactor`\s indexed by name for retrieval when loading objects
        using a Marshmallow schema.

    :returns:
        both 'obj' and 'mentioned', as updated
    """
    if obj.get("name"):
        if obj["name"] in mentioned:
            if obj.get("anchors"):
                for anchor in obj["anchors"]:
                    mentioned[obj["name"]]["anchors"].append(anchor)
        else:
            mentioned.insert_by_name(obj)
        obj = obj["name"]
    return obj, mentioned


def collect_mentioned(
    obj: Union[RawFactor, List[Union[RawFactor, str]]],
    mentioned: Optional[Mentioned] = None,
    keys_to_ignore: Sequence[str] = (
        "predicate",
        "anchors",
        "factor_anchors",
        "enactment_anchors",
    ),
) -> Tuple[RawFactor, Mentioned]:
    """
    Make a dict of all nested objects labeled by name, creating names if needed.
    To be used during loading to expand name references to full objects.
    """
    mentioned = mentioned or Mentioned()
    if isinstance(obj, List):
        new_list = []
        for item in obj:
            new_item, new_mentioned = collect_mentioned(item, mentioned)
            mentioned.update(new_mentioned)
            new_list.append(new_item)
        obj = new_list
    if isinstance(obj, Dict):

        obj, mentioned = update_name_index_from_fact_content(obj, mentioned)

        for key, value in obj.items():
            if key not in keys_to_ignore:
                if isinstance(value, (Dict, List)):
                    new_value, new_mentioned = collect_mentioned(value, mentioned)
                    mentioned.update(new_mentioned)
                    obj[key] = new_value
        obj = ensure_factor_has_name(obj)

        # Added because a top-level factor was not having its brackets replaced
        if obj.get("predicate", {}).get("content"):
            for factor in obj.get("terms", []):
                if factor not in obj["predicate"]["content"]:
                    obj["predicate"][
                        "content"
                    ] = text_expansion.replace_brackets_with_placeholder(
                        content=obj["predicate"]["content"], name=factor
                    )

        obj, mentioned = update_name_index_with_factor(obj, mentioned)
    return obj, mentioned


def index_names(record: Union[List, Dict]) -> Mentioned:
    """
    Call all functions to prepare "mentioned" index.

    The names are sorted by length so that if one mentioned Factor's name
    is a substring of another, the longest available name is expanded.

    :returns:
        a modified version of the dict to load, plus a dict of names
        and the objects to expand them with.
    """
    obj = deepcopy(record)
    obj = text_expansion.expand_shorthand(obj)
    obj, mentioned = collect_mentioned(obj)
    sorted_mentioned = mentioned.sorted_by_length()
    return obj, sorted_mentioned



import pytest

from authorityspoke.holdings import HoldingGroup


class TestMakeHoldingGroup:
    def test_all_members_must_be_holdings(self, make_rule):
        with pytest.raises(TypeError):
            HoldingGroup([make_rule["h1"]])


class TestHoldingGroupImplies:
    def test_explain_holdinggroup_implication(self, make_holding):
        left = HoldingGroup([make_holding["h1"], make_holding["h2_ALL"]])
        right = HoldingGroup([make_holding["h2"]])
        explanation = left.explain_implication(right)
        assert "implies" in str(explanation).lower()

    def test_implication_of_holding(self, make_holding):
        left = HoldingGroup([make_holding["h1"], make_holding["h2_ALL"]])
        right = make_holding["h2"]
        assert left.implies(right)

    def test_implication_of_rule(self, make_holding, make_rule):
        left = HoldingGroup([make_holding["h1"], make_holding["h2_ALL"]])
        right = make_rule["h2"]
        assert left.implies(right)

    def test_implication_of_none(self, make_holding):
        left = HoldingGroup([make_holding["h1"], make_holding["h2_ALL"]])
        right = None
        assert left.implies(right)



import datetime
import os

import pytest

from anchorpoint.textselectors import TextPositionSelector, TextSelectionError
from dotenv import load_dotenv

from legislice.citations import CodeLevel
from legislice.download import Client, LegislicePathError
from legislice.yaml_schemas import ExpandableEnactmentSchema

from authorityspoke.io import loaders
from authorityspoke.io.fake_enactments import FakeClient

load_dotenv()

TOKEN = os.getenv("LEGISLICE_API_TOKEN")


class TestCodes:
    client = Client(api_token=TOKEN)

    def test_cfr_repr(self):
        oracle_dictionary = loaders.load_holdings("holding_oracle.json")
        regulation = oracle_dictionary[10]["enactments_despite"][1]
        schema = ExpandableEnactmentSchema()
        enactment = schema.load(regulation)

        assert enactment.level == CodeLevel.REGULATION
        assert "/cfr/" in repr(enactment)

    @pytest.mark.vcr
    @pytest.mark.parametrize(
        "path, heading, level",
        [
            ("/us/usc", "United States Code", CodeLevel.STATUTE),
            ("/us/const", "United States Constitution", CodeLevel.CONSTITUTION),
            ("/test/acts", "Acts of Australia", CodeLevel.STATUTE),
        ],
    )
    def test_code_title(self, path, heading, level):
        enactment = self.client.read(path)
        assert enactment.heading.startswith(heading)
        assert enactment.level == level

    def test_code_select_text(self, fake_beard_client):

        enactment = fake_beard_client.read("/test/acts/47/1")
        assert enactment.text.startswith("This Act may be cited")

    def test_code_select_text_chapeau(self, fake_beard_client):
        enactment = fake_beard_client.read("/test/acts/47/4")
        enactment.select()
        assert enactment.selected_text().startswith("In this Act, beard means")

    def test_get_bill_of_rights_effective_date(self, fake_usc_client):
        enactment = fake_usc_client.read("/us/const/amendment/V")
        assert enactment.start_date == datetime.date(1791, 12, 15)

    def test_text_interval_constitution_section(self, fake_usc_client):
        enactment = fake_usc_client.read("/us/const/article/I/8/8")

        # The entire 317-character section is selected.
        ranges = enactment.selection.ranges()
        assert ranges[0].start == 0
        assert ranges[0].end == 172

        passage = enactment.get_passage(TextPositionSelector(68, 81))
        assert passage == "…limited Times…"

    def test_text_interval_beyond_end_of_section(self, fake_usc_client):
        """No longer raises an error, just selects to the end of the text."""
        enactment = fake_usc_client.read("/us/const/article/I/8/8")

        selector = TextPositionSelector(68, 400)
        passage = enactment.get_passage(selector)
        assert passage.startswith("…limited Times")
        assert passage.endswith(enactment.text[-10:])

    def test_bad_section(self, fake_usc_client):
        """
        The path is a section that doesn't exist in USC.
        """
        with pytest.raises(LegislicePathError):
            enactment = fake_usc_client.read("/us/usc/article/I/3/7")

    def test_text_interval_bad_selector(self, make_selector, make_response):
        client = FakeClient(responses=make_response)
        enactment = client.read("/us/const/amendment/IV")
        with pytest.raises(TextSelectionError):
            _ = enactment.convert_selection_to_set(make_selector["bad_selector"])



"""
Marshmallow schemas for loading AuthoritySpoke objects from JSON.

Intended for use with machine-generated API responses.
Should be suitable for generating an OpenAPI specification.
"""

from typing import Dict, List, NamedTuple, Optional, Sequence, TypedDict, Type, Union

from marshmallow import Schema, fields, EXCLUDE
from marshmallow import pre_load, post_load
from marshmallow_oneofschema import OneOfSchema

from anchorpoint.textselectors import TextQuoteSelector, TextPositionSelector
from anchorpoint.schemas import SelectorSchema
from legislice import Enactment
from legislice.schemas import EnactmentSchema
from nettlesome.factors import Factor
from nettlesome.schemas import PredicateSchema, EntitySchema, RawFactor

from authorityspoke.decisions import CAPCitation, Decision
from authorityspoke.evidence import Exhibit, Evidence
from authorityspoke.facts import Fact
from authorityspoke.holdings import Holding
from authorityspoke.opinions import Opinion
from authorityspoke.pleadings import Pleading, Allegation
from authorityspoke.procedures import Procedure
from authorityspoke.rules import Rule

RawSelector = Union[str, Dict[str, str]]
RawEnactment = Dict[str, Union[str, List[RawSelector]]]
RawProcedure = Dict[str, Sequence[RawFactor]]
RawRule = Dict[str, Union[RawProcedure, Sequence[RawEnactment], str, bool]]
RawHolding = Dict[str, Union[RawRule, str, bool]]


RawOpinion = Dict[str, str]
RawCAPCitation = Dict[str, str]
RawDecision = Dict[str, Union[str, int, Sequence[RawOpinion], Sequence[RawCAPCitation]]]


class CAPCitationSchema(Schema):
    """Schema for Decision citations in CAP API response."""

    __model__ = CAPCitation
    cite = fields.Str()
    reporter = fields.Str(data_key="type", missing=None)
    case_ids = fields.List(fields.Int(), allow_none=True)

    @post_load
    def make_object(self, data: RawCAPCitation, **kwargs) -> CAPCitation:
        """Load citation."""
        return self.__model__(**data)


class OpinionSchema(Schema):
    """
    Schema for Opinions, of which there may be several in one Decision.

        >>> data = {
        ...     "type": "dissent",
        ...     "author": "Ginsburg",
        ...     "text": "I respectfully dissent."}
        >>> schema = OpinionSchema()
        >>> opinion = schema.load(data)
        >>> opinion.position
        'dissent'
    """

    __model__ = Opinion
    position = fields.Str(data_key="type", missing="majority")
    author = fields.Str(missing="")
    text = fields.Str(missing="")

    @pre_load
    def format_data_to_load(self, data: RawOpinion, **kwargs) -> RawOpinion:
        """Standardize author name before loading object."""
        author = data.get("author") or ""
        data["author"] = author.strip(",:")
        return data

    @post_load
    def make_object(self, data: RawOpinion, **kwargs) -> CAPCitation:
        return self.__model__(**data)


class DecisionSchema(Schema):
    """
    Schema for decisions retrieved from Caselaw Access Project API.

        >>> data = {
        ...     "id": 12083986,
        ...     "url": "https://api.case.law/v1/cases/12083986/",
        ...     "name": "Bellingham executors v. Smith",
        ...     "name_abbreviation": "Bellingham v. Smith",
        ...     "decision_date": "1673-04-29",
        ...     "docket_number": "",
        ...     "first_page": "248",
        ...     "last_page": "248",
        ...     "citations": [
        ...         {
        ...             "type": "official",
        ...             "cite": "1 Rec. Co. Ct. 248"
        ...         }
        ...     ],
        ...     "volume": {
        ...         "url": "https://api.case.law/v1/volumes/32044031754302/",
        ...         "volume_number": "1",
        ...         "barcode": "32044031754302"
        ...     },
        ...     "reporter": {
        ...         "url": "https://api.case.law/v1/reporters/742/",
        ...         "full_name": "Records of the Suffolk county court, 1671-1680",
        ...         "id": 742
        ...     },
        ...     "court": {
        ...         "url": "https://api.case.law/v1/courts/suffolk-cty-ct-2/",
        ...         "name_abbreviation": "Suffolk Cty. Ct.",
        ...         "slug": "suffolk-cty-ct-2",
        ...         "id": 11154,
        ...         "name": "Suffolk County Court"
        ...     },
        ...     "jurisdiction": {
        ...         "id": 4,
        ...         "name_long": "Massachusetts",
        ...         "url": "https://api.case.law/v1/jurisdictions/mass/",
        ...         "slug": "mass",
        ...         "whitelisted": False,
        ...         "name": "Mass."
        ...     },
        ...     "cites_to": [],
        ... }
        >>> schema = DecisionSchema()
        >>> decision = schema.load(data)
        >>> decision.jurisdiction
        'Mass.'
        >>> decision.date.year
        1673
        >>> decision.citations[0].cite
        '1 Rec. Co. Ct. 248'
    """

    __model__ = Decision
    name = fields.Str()
    name_abbreviation = fields.Str(missing=None)
    citations = fields.Nested(CAPCitationSchema, many=True)
    opinions = fields.Nested(OpinionSchema, many=True)
    first_page = fields.Int()
    last_page = fields.Int()
    date = fields.Date(data_key="decision_date")
    court = fields.Str()
    jurisdiction = fields.Str(missing=None)
    id = fields.Int()
    cites_to = fields.Nested(CAPCitationSchema, many=True, missing=list)

    class Meta:
        unknown = EXCLUDE

    @pre_load
    def format_data_to_load(self, data: RawDecision, **kwargs) -> RawDecision:
        """Transform data from CAP API response for loading."""
        if not isinstance(data["court"], str):
            data["court"] = data.get("court", {}).get("slug", "")
        if not isinstance(data["jurisdiction"], str):
            data["jurisdiction"] = data.get("jurisdiction", {}).get("name", "")
        data["opinions"] = (
            data.get("casebody", {}).get("data", {}).get("opinions", [{}])
        )
        data.pop("docket_number", None)
        data.pop("casebody", None)
        data.pop("preview", None)
        data.pop("reporter", None)
        data.pop("volume", None)
        del data["url"]
        data.pop("frontend_url", None)

        # change month to ISO date
        if data.get("decision_date") and data["decision_date"].count("-") == 1:
            data["decision_date"] += "-01"
        return data

    @post_load
    def make_object(self, data, **kwargs):
        return self.__model__(**data)


class FactSchema(Schema):
    """
    Schema for Facts, which may contain arbitrary levels of nesting.

        >>> lotus_entity = {"type": "Entity", "name": "Lotus Development Corporation"}
        >>> menu_entity = {"type": "Entity", "name": "the Lotus menu command hierarchy"}
        >>> fact_data = {
        ...             "predicate": {
        ...                 "content": "$owner registered a copyright covering $work"},
        ...             "terms": [lotus_entity, menu_entity]
        ...         }
        >>> schema = FactSchema()
        >>> fact = schema.load(fact_data)
        >>> str(fact)
        'the fact that <Lotus Development Corporation> registered a copyright covering <the Lotus menu command hierarchy>'
    """

    __model__: Type = Fact
    predicate = fields.Nested(PredicateSchema)
    terms = fields.Nested(lambda: FactorSchema(many=True))
    standard_of_proof = fields.Str(missing=None)
    name = fields.Str(missing=None)
    absent = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)

    @post_load
    def make_object(self, data: RawFactor, **kwargs) -> Fact:
        """Make Fact."""
        return self.__model__(**data)


class ExhibitSchema(Schema):
    """
    Schema for an object that may embody a statement.

        >>> lotus_entity = {"type": "Entity", "name": "Lotus Development Corporation"}
        >>> fact_data = {
        ...    "predicate": {
        ...    "content": "$owner registered a copyright covering $work"},
        ...    "terms": [lotus_entity, {"type": "Entity", "name": "the Lotus menu command hierarchy"}]
        ...         }
        >>> exhibit_data = {
        ...     "form": "certificate of copyright registration",
        ...     "statement": fact_data,
        ...     "statement_attribution": {"name": "Lotus Development Corporation"}
        ...     }
        >>> schema = ExhibitSchema()
        >>> exhibit = schema.load(exhibit_data)
        >>> print(exhibit)
        the certificate of copyright registration attributed to <Lotus Development Corporation>, asserting the fact that <Lotus Development Corporation> registered a copyright covering <the Lotus menu command hierarchy>,
    """

    __model__: Type = Exhibit
    form = fields.Str(missing=None)
    statement = fields.Nested(FactSchema, missing=None)
    statement_attribution = fields.Nested(EntitySchema, missing=None)
    name = fields.Str(missing=None)
    absent = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)

    @post_load
    def make_object(self, data: RawFactor, **kwargs) -> Exhibit:
        """Make Exhibit."""
        return self.__model__(**data)


class PleadingSchema(Schema):
    """
    Schema for a document to link Allegations to.

    >>> lotus_entity = {"name": "Lotus Development Corporation"}
    >>> pleading_data = {"filer": lotus_entity, "generic": False}
    >>> schema = PleadingSchema()
    >>> pleading = schema.load(pleading_data)
    >>> str(pleading)
    'the pleading filed by <Lotus Development Corporation>'
    """

    __model__: Type = Pleading
    filer = fields.Nested(EntitySchema, missing=None)
    name = fields.Str(missing=None)
    absent = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)

    @post_load
    def make_pleading(self, data: RawFactor, **kwargs) -> Pleading:
        return self.__model__(**data)


class AllegationSchema(Schema):
    """Schema for an Allegation of a Fact."""

    __model__: Type = Allegation
    pleading = fields.Nested(PleadingSchema, missing=None)
    statement = fields.Nested(FactSchema, missing=None)
    name = fields.Str(missing=None)
    absent = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)

    @post_load
    def make_allegation(self, data: RawFactor, **kwargs) -> Allegation:
        return self.__model__(**data)


class EvidenceSchema(Schema):
    """
    Schema for an Exhibit and a reference to the Fact it would support.

        >>> lotus_entity = {"type": "Entity", "name": "Lotus Development Corporation"}
        >>> menu_entity = {"type": "Entity", "name": "the Lotus menu command hierarchy"}
        >>> fact_data = {
        ...     "predicate": {
        ...     "content": "$owner registered a copyright covering $work"},
        ...     "terms": [lotus_entity, menu_entity]
        ...     }
        >>> exhibit_data = {
        ...     "form": "certificate of copyright registration",
        ...     "statement": fact_data,
        ...     "statement_attribution": {"name": "Lotus Development Corporation"}
        ...     }
        >>> evidence_data = {
        ...     "exhibit": exhibit_data,
        ...     "to_effect": {
        ...         "predicate": {"content": "$work was copyrightable"},
        ...         "terms": [menu_entity]}
        ... }
        >>> schema = EvidenceSchema()
        >>> evidence = schema.load(evidence_data)
        >>> print(evidence)
        the evidence of the certificate of copyright registration attributed to <Lotus Development Corporation>, asserting the fact that <Lotus Development Corporation> registered a copyright covering <the Lotus menu command hierarchy>, which supports the fact that <the Lotus menu command hierarchy> was copyrightable
    """

    __model__: Type = Evidence
    exhibit = fields.Nested(ExhibitSchema, missing=None)
    to_effect = fields.Nested(FactSchema, missing=None)
    name = fields.Str(missing=None)
    absent = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)

    @post_load
    def make_evidence(self, data: RawFactor, **kwargs) -> Evidence:
        return self.__model__(**data)


class FactorSchema(OneOfSchema):
    """
    Schema that directs data to "one of" the other schemas.

    Must include a `type` field to indicate which subclass schema to use.

        >>> lotus_entity = {"type": "Entity", "name": "Lotus Development Corporation"}
        >>> menu_entity = {"type": "Entity", "name": "the Lotus menu command hierarchy"}
        >>> fact_data = {
        ...             "type": "Fact",
        ...             "predicate": {
        ...                 "content": "$owner registered a copyright covering $work"},
        ...             "terms": [lotus_entity, menu_entity]
        ...         }
        >>> schema = FactorSchema()
        >>> fact = schema.load(fact_data)
        >>> str(fact)
        'the fact that <Lotus Development Corporation> registered a copyright covering <the Lotus menu command hierarchy>'
    """

    __model__: Type = Factor
    type_schemas = {
        "Allegation": AllegationSchema,
        "Entity": EntitySchema,
        "Evidence": EvidenceSchema,
        "Exhibit": ExhibitSchema,
        "Fact": FactSchema,
        "Pleading": PleadingSchema,
    }

    def get_obj_type(self, obj) -> str:
        """Return name of object schema."""
        return obj.__class__.__name__


class ProcedureSchema(Schema):
    """
    Schema for loading a Procedure from data fitting an OpenAPI schema.

    Does not require separate :class:`~authorityspoke.facts.TermSequence` schema.
    :class:`~FactorSchema` will load the equivalent of
    a :class:`~nettlesome.facts.TermSequence` with the parameter ``many=True``.

        >>> api_data = {"type": "Entity", "name": "the Java API"}
        >>> procedure_data = {
        ... "inputs": [{
        ...     "type": "Fact",
        ...     "predicate": {"content": "$work was an original work",
        ...     "truth": False},
        ...     "terms": [api_data]}],
        ... "outputs": [{
        ...     "type": "Fact",
        ...     "predicate": {"content": "$work was copyrightable",
        ...     "truth": False},
        ...     "terms": [api_data]}]}
        >>> schema = ProcedureSchema()
        >>> procedure = schema.load(procedure_data)
        >>> procedure.short_string
        'RESULT: the fact it was false that <the Java API> was copyrightable GIVEN: the fact it was false that <the Java API> was an original work'
    """

    __model__: Type = Procedure
    inputs = fields.Nested(FactorSchema, many=True)
    despite = fields.Nested(FactorSchema, many=True)
    outputs = fields.Nested(FactorSchema, many=True)

    @post_load
    def make_object(self, data, **kwargs):
        return self.__model__(**data)


class RuleSchema(Schema):
    """
    Schema for loading Rules from data fitting an OpenAPI schema.

    >>> api_data = {"type": "Entity", "name": "the Java API"}
    >>> procedure_data = {
    ... "inputs": [{
    ...     "type": "Fact",
    ...     "predicate": {"content": "$work was an original work",
    ...     "truth": False},
    ...     "terms": [api_data]}],
    ... "outputs": [{
    ...     "type": "Fact",
    ...     "predicate": {"content": "$work was copyrightable",
    ...     "truth": False},
    ...     "terms": [api_data]}]}
    >>> rule_data = {"procedure": procedure_data,
    ...     "mandatory": True,
    ...     "enactments": [{
    ...         "node": "/us/usc/t17/s102",
    ...         "start_date": "1990-12-01",
    ...         "text_version": {
    ...         "content": "Copyright protection subsists, in accordance with this title, in original works of authorship fixed in any tangible medium of expression, now known or later developed."},
    ...         "selection": [{"start": 0, "end": 93}]}]}
    >>> schema = RuleSchema()
    >>> rule = schema.load(rule_data)
    >>> print(rule.short_string)
    the Rule that the court MUST SOMETIMES impose the RESULT: the fact it was false that <the Java API> was copyrightable GIVEN: the fact it was false that <the Java API> was an original work GIVEN the ENACTMENT: "Copyright protection subsists, in accordance with this title, in original works of authorship…" (/us/usc/t17/s102 1990-12-01)
    """

    __model__: Type = Rule
    procedure = fields.Nested(ProcedureSchema)
    enactments = fields.Nested(EnactmentSchema, many=True)
    enactments_despite = fields.Nested(EnactmentSchema, many=True)
    mandatory = fields.Bool(missing=False)
    universal = fields.Bool(missing=False)
    name = fields.Str(missing=None)
    generic = fields.Bool(missing=False)

    @post_load
    def make_object(self, data, **kwargs):
        return self.__model__(**data)


class HoldingSchema(Schema):
    """
    Schema for loading Holdings from data fitting an OpenAPI schema.

    >>> api_data = {"type": "Entity", "name": "the Java API"}
    >>> procedure_data = {
    ... "inputs": [{
    ...     "type": "Fact",
    ...     "predicate": {"content": "$work was an original work",
    ...     "truth": False},
    ...     "terms": [api_data]}],
    ... "outputs": [{
    ...     "type": "Fact",
    ...     "predicate": {"content": "$work was copyrightable",
    ...     "truth": False},
    ...     "terms": [api_data]}]}
    >>> rule_data = {"procedure": procedure_data,
    ...     "mandatory": True,
    ...     "enactments": [{
    ...         "node": "/us/usc/t17/s102",
    ...         "start_date": "1990-12-01",
    ...         "text_version": {
    ...         "content": "Copyright protection subsists, in accordance with this title, in original works of authorship fixed in any tangible medium of expression, now known or later developed."},
    ...         "selection": [{"start": 0, "end": 93}]}]}
    >>> holding_data = {
    ...     "rule": rule_data,
    ...     "rule_valid": False}
    >>> schema = HoldingSchema()
    >>> holding = schema.load(holding_data)
    >>> print(holding.short_string)
    the Holding to REJECT the Rule that the court MUST SOMETIMES impose the RESULT: the fact it was false that <the Java API> was copyrightable GIVEN: the fact it was false that <the Java API> was an original work GIVEN the ENACTMENT: "Copyright protection subsists, in accordance with this title, in original works of authorship…" (/us/usc/t17/s102 1990-12-01)
    """

    __model__: Type = Holding
    rule = fields.Nested(RuleSchema)
    rule_valid = fields.Bool(missing=True)
    decided = fields.Bool(missing=True)
    exclusive = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)
    anchors = fields.Nested(SelectorSchema, many=True)

    @post_load
    def make_object(self, data, **kwargs):
        return self.__model__(**data)


class NamedAnchors(NamedTuple):
    name: Factor
    anchors: List[TextQuoteSelector]


class NamedAnchorsSchema(Schema):
    __model__ = NamedAnchors

    name = fields.Nested(FactorSchema)
    anchors = fields.Nested(SelectorSchema, many=True)


class AnchoredEnactments(NamedTuple):
    enactment: Enactment
    anchors: List[TextQuoteSelector]


class AnchoredEnactmentsSchema(Schema):
    __model__ = AnchoredEnactments

    enactment = fields.Nested(EnactmentSchema)
    anchors = fields.Nested(SelectorSchema, many=True)


SCHEMAS = list(Schema.__subclasses__()) + [SelectorSchema, EnactmentSchema]


def get_schema_for_item(classname: str) -> Schema:
    """Find the Marshmallow schema for an AuthoritySpoke object."""
    schemas_for_names = {
        "TextPositionSelector": SelectorSchema,
        "TextQuoteSelector": SelectorSchema,
        "Comparison": PredicateSchema,
        "Predicate": PredicateSchema,
        "Fact": FactSchema,
        "Evidence": EvidenceSchema,
        "Exhibit": ExhibitSchema,
        "Allegation": AllegationSchema,
        "Pleading": PleadingSchema,
        "Holding": HoldingSchema,
        "Rule": RuleSchema,
        "Procedure": ProcedureSchema,
        "Enactment": EnactmentSchema,
        "Decision": DecisionSchema,
        "Opinion": OpinionSchema,
    }
    result = schemas_for_names.get(classname)
    if result is None:
        raise ValueError(f"No schema found for class '{classname}'")
    return result()



"""For expanding text in input JSON into a format Marshmallow can load."""

from nettlesome.predicates import StatementTemplate
from re import findall
from string import Template
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from marshmallow import ValidationError
from slugify import slugify

from authorityspoke.io import nesting


def expand_shorthand(obj: Union[List, Dict]) -> Union[List, Dict[str, Any]]:
    """Traverse dict and expand every kind of pre-loading shorthand."""
    return nesting.walk_tree_and_modify(
        obj=obj, func=expand_node_shorthand, ignore=("predicate",)
    )


def expand_node_shorthand(obj: Dict[str, Any]) -> Dict[str, Any]:
    """Expand shorthand at one node while walking tree of input JSON."""
    for list_field in ("terms", "anchors"):
        if obj.get(list_field) is not None:
            obj = wrap_single_element_in_list(obj, list_field)

    to_nest = ["content", "truth", "sign", "expression"]
    obj = nesting.nest_fields(obj, nest="predicate", eggs=to_nest)

    obj = collapse_known_factors(obj)

    return obj


def collapse_known_factors(obj: Dict):
    """Replace all names of known context factors with placeholder strings."""
    if obj.get("terms"):
        for factor in obj["terms"]:
            if isinstance(factor, str):
                name = factor
            else:
                name = factor.get("name")
            if name:
                if name in obj["predicate"]["content"]:
                    obj["predicate"]["content"] = collapse_name_in_content(
                        obj["predicate"]["content"], name
                    )
                else:
                    obj["predicate"]["content"] = replace_brackets_with_placeholder(
                        obj["predicate"]["content"], name
                    )
    return obj


def replace_brackets_with_placeholder(content: str, name: str):
    """Replace brackets with placeholder to show it is referenced in terms."""
    slug = slugify(text=name, separator="_", replacements=[[" ", "_"]])
    placeholder_slug = "${" + slug + "}"
    if placeholder_slug not in content:
        content = content.replace("{}", "${" + slug + "}", 1)
    return content


def collapse_name_in_content(content: str, name: str):
    """Replace name with placeholder to show it is referenced in terms."""
    slug = slugify(text=name, separator="_", replacements=[[" ", "_"]])
    placeholder_slug = "${" + slug + "}"
    if placeholder_slug not in content:
        content = content.replace(name, placeholder_slug)
        content = content.replace("{${", "${", 1).replace("}}", "}", 1)
    return content


def add_found_context_with_brackets(
    content: str, terms: List[Dict], factor: Dict, placeholder="{}"
) -> Tuple[str, List[Dict[str, Any]]]:
    """
    Remove bracketed mentions of factor with placeholder and list replacements.

    :returns:
        Content with mentions of factor replaced by placeholder, and
        a terms list with the factor inserted in a position
        corresponding to its position in the context phrase.
    """
    bracketed_name = "{" + factor["name"] + "}"
    if bracketed_name in content:
        index_in_content = content.index(bracketed_name)
        index_in_factor_list = content[:index_in_content].count("${")
        terms.insert(index_in_factor_list, factor)
        content = collapse_name_in_content(content, factor["name"])
    return content, terms


def add_found_context(
    content: str, terms: List[Dict], factor: Dict
) -> Tuple[str, List[Dict[str, Any]]]:
    """
    Replace mentions of factor with placeholder and list replacements.

    :returns:
        Content with mentions of factor replaced by placeholder, and
        a terms list with the factor inserted in a position
        corresponding to its position in the context phrase.
    """
    # if factor["name"] is inside an existing placeholder, replace it with a special placeholder
    template = Template(content)
    matches = list(template.pattern.finditer(template.template))
    for match in matches[::-1]:  # backwards to avoid changing index of a later match
        mangled_placeholder = content[match.start() : match.end()].replace(
            factor["name"], "@@@@"
        )
        content = (
            content[: match.start()] + mangled_placeholder + content[match.end() :]
        )

    if factor["name"] in content:
        index_in_content = content.index(factor["name"])
        index_in_factor_list = content[:index_in_content].count("$")
        terms.insert(index_in_factor_list, factor)
        content = collapse_name_in_content(content, factor["name"])

    # replace the special placeholder with factor["name"]
    content = content.replace("@@@@", factor["name"])
    return content, terms


def get_references_from_string(
    content: str, terms: Optional[List[Dict]]
) -> Tuple[str, Sequence[Dict]]:
    r"""
    Make :class:`.Entity` context :class:`.Factor`\s from string.

    This function identifies context :class:`.Factor`\s by finding
    brackets around them, while :func:`get_references_from_mentioned`
    depends on knowing the names of the context factors in advance.
    Also, this function works only when all the terms
    are type :class:`.Entity`.

    Despite "placeholder" being defined as a variable elsewhere,
    this function isn't compatible with any placeholder string other
    than "{}".

    :param content:
        a string containing a clause making an assertion.
        Curly brackets surround the names of :class:`.Entity`
        context factors to be created.

    :returns:
        a :class:`Predicate` and :class:`.Entity` objects
        from a string that has curly brackets around the
        context factors and the comparison/quantity.
    """
    pattern = r"(?<!\$)\{([^\{]+)\}"  # matches bracketed text not preceded by $
    entities_as_text = findall(pattern, content)
    entities_as_text.sort(key=len, reverse=True)
    terms = terms or []
    for entity_name in entities_as_text:
        entity = {"type": "Entity", "name": entity_name}
        content, terms = add_found_context_with_brackets(content, terms, entity)

    return content, terms


def wrap_single_element_in_list(data: Dict, many_element: str):
    """Make a specified field a list if it isn't already a list."""
    if not isinstance(data[many_element], list):
        data[many_element] = [data[many_element]]
    return data



from __future__ import annotations

from collections import OrderedDict
from copy import deepcopy
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from legislice.enactments import RawEnactment

RawPredicate = Dict[str, Union[str, bool]]
RawFactor = Dict[str, Union[RawPredicate, Sequence[Any], str, bool]]


class EnactmentIndex(OrderedDict):
    """Index of cross-referenced objects, keyed to phrases that reference them."""

    def insert_by_name(self, obj: Dict) -> None:
        """Add record to dict, using value of record's "name" field as the dict key."""
        self[obj["name"]] = obj.copy()
        self[obj["name"]].pop("name")
        return None

    def get_by_name(self, name: str) -> Dict:
        """
        Convert retrieved record so name is a field rather than the key for the whole record.
        :param name:
            the name of the key where the record can be found in the Mentioned dict.
        :returns:
            the value stored at the key "name", plus a name field.
        """
        value = {"name": name}
        value.update(self[name])
        return value

    def __repr__(self):
        return f"EnactmentIndex({repr(dict(self))})"

    def enactment_has_anchor(
        self, enactment_name: str, anchor: Dict[str, Union[str, int]]
    ) -> bool:
        anchors_for_selected_element = self[enactment_name].get("anchors") or []
        return any(
            existing_anchor == anchor
            for existing_anchor in anchors_for_selected_element
        )

    def add_anchor_for_enactment(
        self, enactment_name: str, anchor: Dict[str, Union[str, int]]
    ) -> None:
        anchors_for_selected_element = self[enactment_name].get("anchors") or []
        if not self.enactment_has_anchor(enactment_name, anchor):
            anchors_for_selected_element.append(anchor)
        self[enactment_name]["anchors"] = anchors_for_selected_element

    def __add__(self, other: EnactmentIndex) -> EnactmentIndex:
        new_index = deepcopy(self)
        for key in other.keys():
            other_dict = other.get_by_name(key)
            new_index.index_enactment(other_dict)
        return new_index

    def index_enactment(self, obj: RawEnactment) -> Union[str, RawEnactment]:
        r"""
        Update index of mentioned Factors with 'obj', if obj is named.
        If there is already an entry in the mentioned index with the same name
        as obj, the old entry won't be replaced. But if any additional text
        anchors are present in the new obj, the anchors will be added.
        If obj has a name, it will be collapsed to a name reference.
        :param obj:
            data from JSON to be loaded as a :class:`.Enactment`
        """
        if obj.get("name"):
            if obj["name"] in self:
                if obj.get("anchors"):
                    for anchor in obj["anchors"]:
                        self.add_anchor_for_enactment(
                            enactment_name=obj["name"], anchor=anchor
                        )
            else:
                self.insert_by_name(obj)
            obj = obj["name"]
        return obj


def create_name_for_enactment(obj: RawEnactment) -> str:
    name: str = obj["node"]
    if obj.get("start_date"):
        name += f'@{obj["start_date"]}'

    for field_name in ["start", "end", "prefix", "exact", "suffix"]:
        if obj.get(field_name):
            name += f':{field_name}="{obj[field_name]}"'
    return name


def ensure_enactment_has_name(obj: RawEnactment) -> RawEnactment:

    if not obj.get("name"):
        new_name = create_name_for_enactment(obj)
        if new_name:
            obj["name"] = new_name
    return obj


def collect_enactments(
    obj: Union[RawFactor, List[Union[RawFactor, str]]],
    mentioned: Optional[EnactmentIndex] = None,
    keys_to_ignore: Sequence[str] = ("predicate", "anchors", "children"),
) -> Tuple[RawFactor, EnactmentIndex]:
    """
    Make a dict of all nested objects labeled by name, creating names if needed.
    To be used during loading to expand name references to full objects.
    """
    mentioned = mentioned or EnactmentIndex()
    if isinstance(obj, List):
        new_list = []
        for item in obj:
            new_item, new_mentioned = collect_enactments(item, mentioned)
            mentioned.update(new_mentioned)
            new_list.append(new_item)
        obj = new_list
    if isinstance(obj, Dict):
        new_dict = {}
        for key, value in obj.items():
            if key not in keys_to_ignore and isinstance(value, (Dict, List)):
                new_value, new_mentioned = collect_enactments(value, mentioned)
                mentioned.update(new_mentioned)
                new_dict[key] = new_value
            else:
                new_dict[key] = value

        if new_dict.get("node") or (new_dict.get("name") in mentioned.keys()):
            new_dict = ensure_enactment_has_name(new_dict)
            new_dict = mentioned.index_enactment(new_dict)
        obj = new_dict
    return obj, mentioned


"""
This file (one_of_schema.py) is derived from marshmallow-oneofschema.

marshmallow-oneofschema is subject to the following MIT license:

Copyright 2016-2017 Maxim Kulkin
Copyright 2018 Alex Rothberg and contributors

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from typing import Dict, Type
from marshmallow import Schema, ValidationError


class OneOfSchema(Schema):
    """
    Schema that multiplexes other schemas based on object type.

    This is a special kind of schema that actually multiplexes other schemas
    based on object type. When serializing values, it uses get_obj_type() method
    to get object type name. Then it uses `type_schemas` name-to-Schema mapping
    to get schema for that particular object type, serializes object using that
    schema and adds an extra "type" field with name of object type.
    Deserialization is reverse.

    Example:
        class Foo(object):
            def __init__(self, foo):
                self.foo = foo

        class Bar(object):
            def __init__(self, bar):
                self.bar = bar

        class FooSchema(marshmallow.Schema):
            foo = marshmallow.fields.String(required=True)

            @marshmallow.post_load
            def make_foo(self, data, **kwargs):
                return Foo(**data)

        class BarSchema(marshmallow.Schema):
            bar = marshmallow.fields.Integer(required=True)

            @marshmallow.post_load
            def make_bar(self, data, **kwargs):
                return Bar(**data)

        class MyUberSchema(marshmallow.OneOfSchema):
            type_schemas = {
                'foo': FooSchema,
                'bar': BarSchema,
            }

            def get_obj_type(self, obj):
                if isinstance(obj, Foo):
                    return 'foo'
                elif isinstance(obj, Bar):
                    return 'bar'
                else:
                    raise Exception('Unknown object type: %s' % repr(obj))

        MyUberSchema().dump([Foo(foo='hello'), Bar(bar=123)], many=True)
        # => [{'type': 'foo', 'foo': 'hello'}, {'type': 'bar', 'bar': 123}]

    You can control type field name added to serialized object representation by
    setting `type_field` class property.
    """

    type_field = "type"
    type_field_remove = True
    type_schemas: Dict[str, Type[Schema]] = {}

    def get_obj_type(self, obj):
        """Return name of object schema."""
        return obj.__class__.__name__

    def dump(self, obj, *, many=None, **kwargs):
        """Dump object to JSON."""
        errors = {}
        result_data = []
        result_errors = {}
        many = self.many if many is None else bool(many)
        if not many:
            result = result_data = self._dump(obj, **kwargs)
        else:
            for idx, o in enumerate(obj):
                try:
                    result = self._dump(o, **kwargs)
                    result_data.append(result)
                except ValidationError as error:
                    result_errors[idx] = error.normalized_messages()
                    result_data.append(error.valid_data)

        result = result_data
        errors = result_errors

        if not errors:
            return result
        else:
            exc = ValidationError(errors, data=obj, valid_data=result)
            raise exc

    def _dump(self, obj, *, update_fields=True, **kwargs):
        obj_type = self.get_obj_type(obj)
        if not obj_type:
            return (
                None,
                {"_schema": "Unknown object class: %s" % obj.__class__.__name__},
            )

        type_schema = self.type_schemas.get(obj_type)
        if not type_schema:
            return None, {"_schema": "Unsupported object type: %s" % obj_type}

        schema = type_schema if isinstance(type_schema, Schema) else type_schema()

        schema.context.update(getattr(self, "context", {}))

        result = schema.dump(obj, many=False, **kwargs)
        if result is not None:
            result[self.type_field] = obj_type
        return result

    def load(self, data, *, many=None, partial=None, unknown=None):
        """Load object from JSON."""
        errors = {}
        result_data = []
        result_errors = {}
        many = self.many if many is None else bool(many)
        if partial is None:
            partial = self.partial
        if not many:
            try:
                result = result_data = self._load(
                    data, partial=partial, unknown=unknown
                )
                #  result_data.append(result)
            except ValidationError as error:
                result_errors = error.normalized_messages()
                result_data.append(error.valid_data)
        else:
            for idx, item in enumerate(data):
                try:
                    result = self._load(item, partial=partial)
                    result_data.append(result)
                except ValidationError as error:
                    result_errors[idx] = error.normalized_messages()
                    result_data.append(error.valid_data)

        result = result_data
        errors = result_errors

        if not errors:
            return result
        else:
            exc = ValidationError(errors, data=data, valid_data=result)
            raise exc

    def pre_load(self, data, **kwargs):
        """
        Preprocess data before selecting from among the linked schema options.

        To be replaced by a function that would have used the @pre_load decorator,
        if it was available for OneOfSchema objects.
        """

        return data

    def _load(self, data, *, partial=None, unknown=None):
        data = self.pre_load(data)
        if not isinstance(data, dict):
            raise ValidationError({"_schema": "Invalid data type: %s" % data})

        data = dict(data)
        unknown = unknown or self.unknown

        data_type = data.get(self.type_field)
        if self.type_field in data and self.type_field_remove:
            data.pop(self.type_field)

        if not data_type:
            raise ValidationError(
                {self.type_field: ["Missing data for required field."]}
            )

        try:
            type_schema = self.type_schemas.get(data_type)
        except TypeError:
            # data_type could be unhashable
            raise ValidationError({self.type_field: ["Invalid value: %s" % data_type]})
        if not type_schema:
            raise ValidationError(
                {self.type_field: ["Unsupported value: %s" % data_type]}
            )

        schema = type_schema if isinstance(type_schema, Schema) else type_schema()

        schema.context.update(getattr(self, "context", {}))

        return schema.load(data, many=False, partial=partial, unknown=unknown)

    def validate(self, data, *, many=None, partial=None):
        """Check if loading data produces error messages."""
        try:
            self.load(data, many=many, partial=partial)
        except ValidationError as ve:
            return ve.messages
        return {}



r"""
:class:`Court` documents that decide litigation and posit :class:`.Rule`\s.

Unlike most other ``authorityspoke`` classes, :class:`Opinion`\s are not frozen.
"""

from __future__ import annotations

from itertools import zip_longest
from typing import Any, Dict, Iterable, Iterator, List, NamedTuple
from typing import Optional, Sequence, Union

import logging
import re

from dataclasses import dataclass, field

from anchorpoint.textselectors import TextQuoteSelector

from nettlesome.terms import Comparable, ContextRegister, Explanation
from nettlesome.factors import Factor
from authorityspoke.holdings import Holding, HoldingGroup
from authorityspoke.procedures import Procedure
from authorityspoke.rules import Rule


logger = logging.getLogger(__name__)


TextLinkDict = Dict[str, List[TextQuoteSelector]]


class AnchoredHoldings(NamedTuple):
    """Holdings with objects storing the Holdings' links to Opinion text."""

    holdings: List[Holding]
    holding_anchors: List[List[TextQuoteSelector]]
    named_anchors: TextLinkDict
    enactment_anchors: TextLinkDict


@dataclass
class Opinion(Comparable):
    """
    A document that resolves legal issues in a case and posits legal holdings.

    Usually an opinion must have ``position="majority"``
    to create holdings binding on any courts.

    :param position:
        the opinion's attitude toward the court's disposition of the case.
        e.g. ``majority``, ``dissenting``, ``concurring``, ``concurring in the result``
    :param author:
        name of the judge who authored the opinion, if identified
    :param text:
    """

    position: str = "majority"
    author: Optional[str] = None
    text: Optional[str] = field(default=None, repr=False)

    def __post_init__(self):
        r"""
        Add attributes to store Holdings and Factors keyed to their text selectors.

        The ``holding_anchors`` are text selectors for the whole :class:`Holding`, not for any
        individual :class:`.Factor`. Often select text used to
        indicate whether the :class:`.Rule` is ``mandatory``, ``universal``,
        ``valid``, or ``decided``, or show the ``exclusive`` way to reach
        the outputs.
        """

        self._holdings = HoldingGroup()
        self.factor_anchors = {}
        self.enactment_anchors = {}

    def factors(self) -> List[Factor]:
        factors_by_name = self.factors_by_name()
        return list(factors_by_name.values())

    def factors_by_name(self) -> FactorIndex:
        factor_index = FactorIndex()
        for holding in self.holdings:
            factors_for_holding = holding.recursive_terms
            for value in factors_for_holding.values():
                if not isinstance(value, Holding):
                    factor_index.insert_by_name(value=value)
        return factor_index

    @property
    def holdings(self):
        return HoldingGroup(self._holdings)

    def clear_holdings(self):
        r"""Remove all :class:`.Holding`\s from the opinion."""
        self._holdings = HoldingGroup()

    def explanations_contradiction(
        self,
        other: Comparable,
    ) -> Iterator[Explanation]:
        """Yield contexts that would result in a contradiction between self and other."""
        if not self.comparable_with(other):
            raise TypeError(
                f"'Implies' test not implemented for types {self.__class__} and {other.__class__}."
            )
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, Holding):
            yield from self.holdings.explanations_contradiction(other)

        elif isinstance(other, self.__class__):
            yield from self.holdings.explanations_contradiction(other.holdings)

        elif hasattr(other, "explanations_contradiction"):
            yield from other.explanations_contradiction(self)

    def comparable_with(self, other: Any) -> bool:
        """Check if other can be compared to self for implication or contradiction."""
        if not isinstance(other, Comparable):
            return False
        if isinstance(other, Procedure):
            return False
        return not isinstance(other, Factor)

    def contradicts(
        self, other: Comparable, context: Optional[ContextRegister] = None
    ) -> bool:
        if not self.comparable_with(other):
            raise TypeError(
                "'Contradicts' test not implemented for types "
                + f"{self.__class__} and {other.__class__}."
            )
        return any(
            explanation is not None
            for explanation in self.explanations_contradiction(other)
        )

    def implies(
        self, other: Optional[Comparable], context: Optional[ContextRegister] = None
    ) -> bool:
        if other is None:
            return True
        if isinstance(other, HoldingGroup):
            return self.implies_other_holdings(other)
        return any(self.explanations_implication(other))

    def explain_implication(
        self,
        other: Comparable,
    ) -> Optional[Explanation]:
        explanations = self.explanations_implication(other)
        try:
            explanation = next(explanations)
        except StopIteration:
            return None
        return explanation

    def explain_contradiction(self, other: Comparable) -> Optional[Explanation]:
        explanations = self.explanations_contradiction(other)
        try:
            explanation = next(explanations)
        except StopIteration:
            return None
        return explanation

    def explanations_implication(
        self,
        other: Comparable,
    ) -> Iterator[Union[ContextRegister, Explanation]]:
        """Yield contexts that would result in self implying other."""
        if not self.comparable_with(other):
            raise TypeError(
                f"'Implies' test not implemented for types {self.__class__} and {other.__class__}."
            )
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, Holding):
            for self_holding in self.holdings:
                yield from self_holding.explanations_implication(other)
        if isinstance(other, HoldingGroup):
            yield from self.holdings.explanations_implication(other)
        elif isinstance(other, self.__class__):
            yield from self.holdings.explanations_implication(other.holdings)
        else:
            yield from other.explanations_implied_by(self)

    def generic_terms_by_str(self) -> Dict[str, Comparable]:
        r"""
        Get all generic :class:`.Factor`\s mentioned in ``self``.

        :returns:
            a list of generic :class:`.Factor` objects mentioned in
            any ``input``, ``output``, or ``despite`` :class:`.Factor`
            of ``self``, with guaranteed order, including each
            generic :class:`.Factor` only once.
        """
        generics: Dict[str, Comparable] = {}
        for holding in self.holdings:
            for generic in holding.generic_terms():
                generics[str(generic)] = generic
        return generics

    def get_factor_by_name(self, name: str) -> Optional[Comparable]:
        """
        Search recursively in holdings of ``self`` for :class:`.Factor` with ``name``.

        :param name:
            string to match with the ``name`` attribute of the
            :class:`.Factor` being searched for.

        :returns:
            a :class:`.Factor` with the specified ``name``, if one
            exists in a :class:`.Rule` in ``self.holdings``.
            Otherwise ``None``.
        """

        for holding in self.holdings:
            factor = holding.get_factor_by_name(name)
            if factor is not None:
                return factor
        return None

    def get_factor_by_str(self, query: str) -> Optional[Factor]:
        r"""
        Search recursively  for :class:`.Factor` in holdings of self.
        """
        return self.holdings.get_factor_by_str(query)

    def get_matching_holding(self, holding: Holding) -> Optional[Holding]:
        for known_holding in self.holdings:
            if holding.means(known_holding):
                return known_holding
        return None

    def posit_holding(
        self,
        holding: Union[Holding, Rule],
        holding_anchors: Optional[
            Union[TextQuoteSelector, List[TextQuoteSelector]]
        ] = None,
        named_anchors: Optional[TextLinkDict] = None,
        enactment_anchors: Optional[TextLinkDict] = None,
        context: Optional[Sequence[Factor]] = None,
    ) -> None:
        r"""Record that this Opinion endorses specified :class:`Holding`\s."""
        if isinstance(holding, Rule):
            logger.warning(
                "posit_holding was called with a Rule "
                "that was automatically converted to a Holding"
            )
            holding = Holding(rule=holding)

        if not isinstance(holding, Holding):
            raise TypeError('"holding" must be an object of type Holding.')

        if holding_anchors and not isinstance(holding_anchors, List):
            holding_anchors = [holding_anchors]

        if holding_anchors:
            holding.anchors = holding.anchors + holding_anchors

        if named_anchors:
            self.factor_anchors = {**self.factor_anchors, **named_anchors}

        if enactment_anchors:
            self.enactment_anchors = {**self.enactment_anchors, **enactment_anchors}

        matching_holding = self.get_matching_holding(holding)
        if matching_holding:
            matching_holding.anchors += holding.anchors
        else:
            if context:
                holding = holding.new_context(context, source=self)
            self._holdings = self._holdings + HoldingGroup(holding)

    def posit_holdings(
        self,
        holdings: Union[AnchoredHoldings, Iterable[Union[Holding, Rule]]],
        holding_anchors: Optional[List[List[TextQuoteSelector]]] = None,
        named_anchors: Optional[TextLinkDict] = None,
        enactment_anchors: Optional[TextLinkDict] = None,
        context: Optional[Sequence[Factor]] = None,
    ):
        r"""
        Add :class:`.Holding`\s to this ``Opinion`` from a sequence.

        :param holdings:
            a sequence of :class:`.Holding`\s that this :class:`.Opinion`
            posits as valid in its own court or jurisdiction, regardless of
            whether ``self`` accepts that the ``inputs`` correspond to the
            reality of the current case, and regardless of whether the
            court orders that the ``outputs`` be put into effect.

        :param text_links:
            mapping of :class:`Factor`\s to the :class:`Opinion` passages where
            they can be found. Can be obtained as the `mentioned` return value
            of one of the functions in :mod:`authorityspoke.io.readers`\.

        :param context:
            an ordered sequence (probably :py:class:`dict`) of
            generic :class:`.Factor` objects from ``self`` which
            will provide the context for the new holding in the
            present case.
        """
        if isinstance(holdings, AnchoredHoldings):
            (holdings, holding_anchors, named_anchors, enactment_anchors) = holdings

        holding_anchors = holding_anchors or []
        for (holding, selector_list) in zip_longest(holdings, holding_anchors):
            self.posit_holding(
                holding,
                holding_anchors=selector_list,
                named_anchors=named_anchors,
                enactment_anchors=enactment_anchors,
                context=context,
            )

    def posit(
        self,
        holdings: Union[AnchoredHoldings, Holding, Iterable[Union[Holding, Rule]]],
        holding_anchors: Optional[
            List[Union[TextQuoteSelector, List[TextQuoteSelector]]]
        ] = None,
        named_anchors: Optional[TextLinkDict] = None,
        enactment_anchors: Optional[TextLinkDict] = None,
        context: Optional[Sequence[Factor]] = None,
    ) -> None:
        r"""
        Add one or more :class:`.Holding`\s to this ``Opinion``.

        This method passes its values to :meth:`~posit_holding` or
        :meth:`~posit_holdings` depending on whether the `holding` parameter
        is one :class:`.Holding` or a :class:`list`\.

        :param holdings:
            a :class:`.Holding` that the :class:`.Opinion` ``self`` posits
            as valid in its own court or jurisdiction, regardless of
            whether ``self`` accepts that the ``inputs`` of the
            :class:`.Holding` correspond to the reality of the current
            case, and regardless of whether the court orders that
            the ``outputs`` of the :class:`.Holding` be put into effect.

        :param text_links:
            list of lists of :class:`.Opinion` passages where references to each
            :class:`.Holding` can be found.

        :param context:
            an ordered sequence (probably :py:class:`dict`) of
            generic :class:`.Factor` objects from ``self`` which
            will provide the context for the new holding in the
            present case.
        """
        if isinstance(holdings, Iterable):
            self.posit_holdings(
                holdings,
                holding_anchors=holding_anchors,
                named_anchors=named_anchors,
                enactment_anchors=enactment_anchors,
                context=context,
            )
        else:
            self.posit_holding(
                holdings,
                holding_anchors=holding_anchors,
                named_anchors=named_anchors,
                enactment_anchors=enactment_anchors,
                context=context,
            )

    def _implied_by_holding(self, other: Holding, context: Explanation) -> bool:
        return all(
            other.implies(self_holding, context=context.reversed_context())
            for self_holding in self.holdings
        )

    def _implied_by_rule(self, other: Rule, context: Explanation) -> bool:
        return self._implied_by_holding(other=Holding(other), context=context)

    def implied_by(
        self,
        other: Union[Opinion, Holding, Rule],
        context: Optional[Union[ContextRegister, Explanation]] = None,
    ) -> bool:
        """Determine if other implies all the Holdings of self."""
        if not isinstance(context, Explanation):
            context = Explanation.from_context(context)
        if isinstance(other, Holding):
            return self._implied_by_holding(other, context=context)
        elif isinstance(other, Rule):
            return self._implied_by_rule(other, context=context)
        return other.implies(self, context=context.reversed_context())

    def implies_other_holdings(self, other_holdings: Sequence[Holding]):
        for other_holding in other_holdings:
            if not any(
                self_holding.implies(other_holding) for self_holding in self.holdings
            ):
                return False
        return True

    def __ge__(self, other: Union[Opinion, Rule]) -> bool:
        """
        Find whether ``self``'s holdings imply all the holdings of ``other``.

        :returns:
            a bool indicating whether the :class:`.Rule` ``other``
            (or every holding of ``other``, if other is an :class:`.Opinion`)
            is implied by some :class:`.Rule` in ``self.holdings``.
        """
        return self.implies(other)

    def __gt__(self, other) -> bool:
        """
        Find whether ``self``\'s holdings imply ``other``\'s but self != other.

        This actually tests for inequality because ``Opinion`` does not
        have a ``means`` method.

        :returns:
            self >= other and self != other.
        """
        return (self >= other) and (self != other)

    def select_text(self, selector: TextQuoteSelector) -> Optional[str]:
        r"""
        Get text using a :class:`.TextQuoteSelector`.

        :param selector:
            a selector referencing a text passage in this :class:`Opinion`.

        :returns:
            the text referenced by the selector, or ``None`` if the text
            can't be found.
        """
        if re.search(selector.passage_regex(), self.text, re.IGNORECASE):
            return selector.exact
        raise ValueError(
            f'Passage "{selector.exact}" from TextQuoteSelector '
            + f'not found in Opinion "{self}".'
        )

    def __str__(self):
        return f"{self.position} Opinion by {self.author}"


class FactorIndex(Dict[str, Factor]):
    r"""Index of :class:`.Factor`/s that may share common anchors."""

    def insert_by_name(self, value: Factor) -> None:
        """
        Insert Factor using its name as key if possible.

        If the Factor has no name attr, use its str as key instead.
        """
        if value.name:
            self.insert(key=value.name, value=value)
            return None
        key = str(value)
        self.insert(key=key, value=value)

    def insert(self, key: str, value: Factor) -> None:
        """Insert Factor using its str as its key."""
        if key in self.keys():
            if value.name:
                if not self[key].name:
                    self[key].name = value.name
                if value.name != self[key].name:
                    raise NameError(
                        f"{type(value)} objects with identical representation ({str(value)}) "
                        f"have differing names: '{value.name}' and '{self[key].name}'"
                    )
        else:
            self[key] = value



r"""
Descriptions of procedural changes in litigation.

Does not specify whether they are mandatory or universal,
or specify the :class:`.Enactment`\s that might require them.
"""

from __future__ import annotations
from copy import deepcopy

from itertools import chain
import operator

from typing import Any, Dict, Iterable, Iterator
from typing import List, Optional, Sequence, Tuple, Union

from nettlesome.terms import (
    Comparable,
    ContextRegister,
    Explanation,
    new_context_helper,
    Term,
    TermSequence,
)
from nettlesome.factors import Factor
from nettlesome.groups import FactorGroup
from nettlesome.formatting import indented


class Procedure(Comparable):
    r"""
    A (potential) rule for courts to use in resolving litigation.

    Described in terms of inputs and outputs, and also potentially
    "despite" :class:`.Factor`\s, which occur when a :class:`Rule`
    could be said to apply "even if" some "despite" factor is true.

    Users generally should not need to interact with this class
    directly, under the current design. Instead, they should interact
    with the class :class:`.Rule`.

    :param outputs:
        an outcome that a court may accept based on the presence
        of the ``inputs``

    :param inputs:
        supporting :class:`.Factor`\s in favor of the ``output``.
        The ``input`` :class:`.Factor`\s are not treated as
        potential undercutters.

    :param despite:
        :class:`.Factor`\s that do not prevent the court from
        imposing the ``output``. These could be considered
        "failed undercutters" in defeasible logic. If a :class:`.Factor`
        is relevant both as support for the output and as
        a potential undercutter, include it in both ``inputs``
        and ``despite``.

    :param name:
        An identifier that can be used to reference and
        incorporate an existing procedure in a new
        :class:`.Factor`, instead of constructing a new
        copy of the :class:`.Procedure`.

    :param absent:
        Whether the absence, rather than presence, of this
        :class:`.Procedure` is being referenced. The usefulness
        of this is unclear, but I'm not prepared to eliminate it
        before the :class:`.Argument` class has been implemented
        and tested.

    :param generic:
        Whether the this :class:`Procedure` is being referenced
        as a generic example, which could be replaced by any
        other :class:`Procedure`.
    """

    def __init__(
        self,
        outputs: FactorGroup,
        inputs: Optional[FactorGroup] = None,
        despite: Optional[FactorGroup] = None,
        generic: bool = False,
        absent: bool = False,
        name: str = "",
    ):
        self.outputs = FactorGroup(outputs)
        self.inputs = FactorGroup(inputs) if inputs else FactorGroup()
        self.despite = FactorGroup(despite) if despite else FactorGroup()
        self.name = name
        self.absent = False
        self.generic = False
        self.context_factor_names = ("outputs", "inputs", "despite")

    @property
    def groups(self) -> List[FactorGroup]:
        return [self.outputs, self.inputs, self.despite]

    def add(
        self,
        other: Comparable,
        context: Optional[Union[ContextRegister, Explanation]] = None,
    ) -> Optional[Procedure]:
        """Show how first Procedure triggers the second if not both are universal."""
        if not isinstance(other, self.__class__):
            return self.with_factor(other)
        for explanation in self.triggers_next_procedure(other, context=context):
            added = self._trigger_addition(other, explanation)
            if added:
                return added
        return None

    def __add__(self, other: Procedure) -> Optional[Procedure]:
        """Show how first Procedure triggers the second if not both are universal."""
        return self.add(other)

    def _add_if_universal(
        self, other: Procedure, explanation: Explanation
    ) -> Optional[Procedure]:
        """Show how first Procedure triggers the second if both are universal."""
        self_output_or_input = FactorGroup((*self.outputs, *self.inputs))
        other_input = list(other.inputs)
        implied_inputs = []
        not_implied = []

        while other_input:
            current = other_input.pop()
            implying_explanation = self_output_or_input.explain_implication(
                current, context=explanation
            )
            if implying_explanation is not None:
                explanation = implying_explanation
                implied_inputs.append(current)
            else:
                not_implied.append(current)

        if not any(implied_inputs):
            return None
        to_combine = Procedure(
            inputs=not_implied, outputs=other.outputs, despite=other.despite
        )
        return self.union(to_combine, context=explanation)

    def _trigger_addition(
        self, other: Procedure, explanation: Explanation
    ) -> Procedure:
        """Add two Procedures, given that they have already been found to be addable."""
        triggered_procedure = other.new_context(explanation.context.reversed())
        new_outputs = [*self.outputs, *triggered_procedure.outputs]
        unique_new_outputs = {}
        for key in new_outputs:
            unique_new_outputs[str(key)] = key
        result = deepcopy(self)
        result.set_outputs(list(unique_new_outputs.values()))
        return result

    def _explanations_union_partial(
        self, other: Procedure, context: Optional[ContextRegister] = None
    ) -> Iterable[ContextRegister]:
        """Yield as much of the context as seems likely correct based on this Procedure."""
        yield from self.likely_contexts(other, context)

    def explanations_union(
        self,
        other: Procedure,
        context: Optional[Union[ContextRegister, Explanation]] = None,
    ) -> Iterator[ContextRegister]:
        if not isinstance(context, Explanation):
            context = Explanation.from_context(context)
        for partial in self._explanations_union_partial(other, context.context):
            for guess in self.possible_contexts(other, partial):
                answer = self._union_from_explanation(other, guess)
                if answer:
                    yield guess

    def _union_from_explanation(
        self, other: Procedure, context: ContextRegister
    ) -> Optional[Procedure]:
        r"""
        Combine two :class:`Procedure`\s into one.

        The new :class:`Procedure` will have all of the ``inputs``, ``outputs``,
        and ``despite`` :class:`.Factor`\s of both ``self`` and ``other``.

        All of the context :class:`.Factor`\s of ``self`` will
        remain the same.
        """

        new_inputs = self.inputs._union_from_explanation(other.inputs, context)
        new_outputs = self.outputs._union_from_explanation(other.outputs, context)
        new_despite = self.despite._union_from_explanation_allow_contradiction(
            other.despite, context
        )

        if any(group is None for group in (new_outputs, new_inputs, new_despite)):
            return None
        return Procedure(outputs=new_outputs, inputs=new_inputs, despite=new_despite)

    def __len__(self):
        r"""
        Get number of generic :class:`.Factor`\s specified for ``self``.

        :returns:
            the number of generic :class:`.Factor`\s that need to be
            specified for this :class:`Procedure`.
        """

        return len(self.generic_terms())

    def __repr__(self):
        text = (
            f"{self.__class__.__name__}(outputs=("
            + f"{', '.join(repr(factor) for factor in self.outputs)}), "
            + f"inputs=({', '.join(repr(factor) for factor in self.inputs)}), "
            + f"despite=({', '.join(repr(factor) for factor in self.despite)}))"
        )
        return text

    def __str__(self):

        text = "RESULT:"
        for f in self.outputs:
            text += "\n" + indented(f.wrapped_string)
        if self.inputs:
            text += "\nGIVEN:"
            for f in self.inputs:
                text += "\n" + indented(f.wrapped_string)
        if self.despite:
            text += "\nDESPITE:"
            for f in self.despite:
                text += "\n" + indented(f.wrapped_string)

        return text

    @property
    def factors_all(self) -> List[Factor]:
        r"""
        Get :class:`.Factor`\s in ``inputs``, ``outputs``, and ``despite``.

        :returns:
            a :class:`list` of all :class:`.Factor`\s.
        """

        inputs = self.inputs or ()
        despite = self.despite or ()
        return [*self.outputs, *inputs, *despite]

    @property
    def recursive_terms(self) -> Dict[str, Term]:
        r"""
        Collect `self`'s :attr:`terms`, and their :attr:`terms`, recursively.

        :returns:
            a :class:`dict` (instead of a :class:`set`,
            to preserve order) of :class:`Term`\s.
        """
        answers: Dict[str, Term] = {}

        for context in self.groups:
            if context:
                answers.update(context.recursive_terms)
        return answers

    @property
    def terms(self) -> TermSequence:
        r"""
        Get :class:`Factor`\s used in comparisons with other :class:`Factor`\s.

        :returns:
            a tuple of attributes that are designated as the ``terms``
            for whichever subclass of :class:`Factor` calls this method. These
            can be used for comparing objects using :meth:`consistent_with`
        """
        result: List[Term] = []
        result.extend(self.outputs)
        result.extend(self.inputs)
        result.extend(self.despite)
        return TermSequence(result)

    def generic_terms_by_str(self) -> Dict[str, Term]:
        r"""
        :class:`.Factor`\s that can be replaced without changing ``self``\s meaning.

        self.generic can't be True for :class:`.Procedure`.

        :returns:
            ``self``'s generic :class:`.Factor`\s,
            which must be matched to other generic :class:`.Factor`\s to
            perform equality or implication tests between :class:`.Factor`\s
            with :meth:`.Factor.means` or :meth:`.Factor.__ge__`.
        """
        generic_dict = {
            str(generic): generic
            for factor in self.factors_all
            for generic in factor.generic_terms()
        }
        return generic_dict

    def add_factor(self, incoming: Factor) -> None:
        """
        Add an input :class:`.Factor`.

        :param incoming:
            the new :class:`.Factor` to be added to input

        :returns:
            None
        """
        new_factors = self.inputs.add_or_raise_error(incoming)
        self.set_inputs(new_factors)

    def with_factor(self, incoming: Factor) -> Optional[Procedure]:
        """
        Create new Procedure with added input :class:`.Factor`.

        :param incoming:
            the new :class:`.Factor` to be added to input

        :returns:
            a new version of ``self`` with the specified change
        """
        new_factors = self.inputs + incoming
        if new_factors is None:
            return None
        result = deepcopy(self)
        result.set_inputs(new_factors)
        return result

    def contradicts(self, other, context: Optional[ContextRegister] = None) -> bool:
        r"""
        Find if ``self`` applying in some cases implies ``other`` cannot apply in some.

        Raises an error because, by analogy with :meth:`.Procedure.implies`\,
        users might expect this method to return ``True`` only when
        :class:`Rule` with ``universal=False`` and Procedure ``self``
        would contradict another :class:`Rule` with ``universal=False``
        and Procedure ``other``. But that would never happen.
        """
        raise NotImplementedError(
            "Procedures do not contradict one another unless one of them ",
            "applies in 'ALL' cases. Consider using ",
            "'Procedure.contradicts_some_to_all' or 'Rule.contradicts'.",
        )

    def contradicts_some_to_all(
        self, other: Procedure, context: Optional[ContextRegister] = None
    ) -> bool:
        r"""
        Find if ``self`` applying in some cases implies ``other`` cannot apply in all.

        :returns:
            whether the assertion that ``self`` applies in
            **some** cases contradicts that ``other`` applies in **all**
            cases, where at least one of the :class:`.Rule`\s is ``mandatory``.
        """

        if not isinstance(other, self.__class__):
            return False
        return any(
            context is not None
            for context in self.explain_contradiction_some_to_all(other, context)
        )

    def _has_input_or_despite_factors_implied_by_all_inputs_of(
        self,
        other: Procedure,
        context: Explanation,
    ) -> Iterator[Explanation]:
        """Check if every input of other implies some input or despite factor of self."""
        self_despite_or_input = FactorGroup((*self.despite, *self.inputs))
        yield from self_despite_or_input._explanations_implied_by(
            other.inputs, explanation=context
        )

    def _has_input_or_despite_factors_implying_all_inputs_of(
        self,
        other: Procedure,
        context: Explanation,
    ) -> Iterator[Explanation]:
        """Check if every input of other is implied by some input or despite factor of self."""
        self_despite_or_input = FactorGroup((*self.despite, *self.inputs))
        yield from self_despite_or_input._explanations_implication(
            other.inputs, explanation=context
        )

    def explain_contradiction_some_to_all(
        self,
        other: Procedure,
        context: Optional[Union[ContextRegister, Explanation]] = None,
    ) -> Iterator[Explanation]:
        """Explain why ``other`` can't apply in all cases if ``self`` applies in some."""
        if not isinstance(context, Explanation):
            context = Explanation.from_context(context)

        # For self to contradict other, either every input of other
        # must imply some input or despite factor of self...
        implied_contexts = self._has_input_or_despite_factors_implied_by_all_inputs_of(
            other, context
        )

        # or every input of other must be implied by
        # some input or despite factor of self.
        implying_contexts = self._has_input_or_despite_factors_implying_all_inputs_of(
            other, context
        )

        # For self to contradict other, some output of other
        # must be contradicted by some output of self.
        seen_contexts = []
        for m in chain(implying_contexts, implied_contexts):
            if m.context not in seen_contexts:
                seen_contexts.append(m.context)
                yield from self.outputs._explanations_contradiction(other.outputs, m)

    def _explain_implication_all_to_all_of_procedure(
        self, other: Procedure, context: ContextRegister
    ) -> Iterator[ContextRegister]:
        yield from self.explanations_same_meaning(other, context)

        def other_outputs_implied(context: Optional[ContextRegister]):
            for explanation in self.outputs.explanations_implication(
                other.outputs, context=context
            ):
                yield explanation

        def self_inputs_implied(explanations: Iterable[ContextRegister]):
            for explanation in explanations:
                for result in other.inputs.explanations_implication(
                    self.inputs, context=explanation
                ):
                    yield result

        for explanation in self_inputs_implied(other_outputs_implied(context)):
            for result in self.inputs.explanations_consistent_with(
                other=other.despite, context=explanation
            ):
                yield result

    def explain_implication_all_to_all(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        """Yield contexts establishing that if self is always valid, other is always valid."""
        context = context or ContextRegister()
        if isinstance(other, self.__class__):
            yield from self._explain_implication_all_to_all_of_procedure(other, context)

    def implies_all_to_all(
        self, other: Procedure, context: Optional[ContextRegister] = None
    ) -> bool:
        """
        Find if ``self`` applying in all cases implies ``other`` applies in all.

        ``self`` does not imply ``other`` if any output of ``other``
        is not equal to or implied by some output of ``self``.

        For ``self`` to imply ``other``, every input of ``self``
        must be implied by some input of ``other``.

        ``self`` does not imply ``other`` if any despite of ``other``
        :meth:`~.Factor.contradicts` an input of ``self``.

        :returns:
            whether the assertion that ``self`` applies in **all** cases
            implies that ``other`` applies in **all** cases.
        """
        return any(
            context is not None
            for context in self.explain_implication_all_to_all(other, context)
        )

    def _explain_implication_of_procedure_all_to_some(
        self, other: Procedure, context: ContextRegister
    ) -> Iterator[ContextRegister]:
        yield from self.explain_implication_all_to_all(other, context)

        other_despite_or_input = FactorGroup((*other.despite, *other.inputs))
        self_despite_or_input = FactorGroup((*self.despite, *self.inputs))

        def other_outputs_implied(context: ContextRegister):
            for explanation in self.outputs.explanations_implication(
                other.outputs, context=context
            ):
                yield explanation

        def other_despite_implied(explanations: Iterator[ContextRegister]):
            for explanation in explanations:
                for result in self_despite_or_input.explanations_implication(
                    other.despite, context=explanation
                ):
                    yield result

        for explanation in other_despite_implied(other_outputs_implied(context)):
            if self.inputs.consistent_with(
                other_despite_or_input, context=explanation.context
            ):
                yield explanation

    def explain_implication_all_to_some(
        self,
        other: Factor,
        context: Optional[Union[ContextRegister, Explanation]] = None,
    ) -> Iterator[Explanation]:
        """Yield contexts establishing that if self is always valid, other is sometimes valid."""
        if not isinstance(context, Explanation):
            context = Explanation.from_context(context)
        if isinstance(other, self.__class__):
            yield from self._explain_implication_of_procedure_all_to_some(
                other=other, context=context
            )

    def implies_all_to_some(
        self, other: Procedure, context: Optional[ContextRegister] = None
    ) -> bool:
        r"""
        Find if ``self`` applying in all cases implies ``other`` applies in some.

        For ``self`` to imply ``other``, every output of ``other``
        must be equal to or implied by some output of ``self``.

        For ``self`` to imply ``other``, every input of ``self`` must not be
        contradicted by any input or despite of ``other``.

        ``self`` does not imply ``other`` if any "despite" :class:`.Factor`\s
        of ``other`` are not implied by inputs of ``self``.

        :returns:
            whether the assertion that ``self`` applies in **all** cases
            implies that ``other`` applies in **some** cases (that is, if
            the list of ``self``'s inputs is not considered an exhaustive list
            of the circumstances needed to invoke the procedure).
        """
        return any(
            context is not None
            for context in self.explain_implication_all_to_some(other, context)
        )

    def _implies_procedure_if_present(
        self, other: Procedure, context: Explanation
    ) -> Iterator[Explanation]:
        def other_outputs_implied(context: Explanation):
            yield from self.outputs.explanations_implication(
                other.outputs, context=context
            )

        def other_inputs_implied(context: Explanation):
            yield from self.inputs.explanations_implication(
                other.inputs, context=context
            )

        def other_despite_implied(context: Explanation):
            despite_or_input = FactorGroup((*self.despite, *self.inputs))
            yield from despite_or_input.explanations_implication(
                other.despite,
                context=context,
            )

        for outputs_explanation in other_outputs_implied(context):
            for inputs_explanation in other_inputs_implied(outputs_explanation):
                for despite_explanation in other_despite_implied(inputs_explanation):
                    yield despite_explanation

    def _implies_if_present(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        r"""
        Find if ``self`` would imply ``other`` if they aren't absent.

        When ``self`` and ``other`` are included in
        :class:`Rule`\s that both apply in **some** cases:

        ``self`` does not imply ``other`` if any output of ``other``
        is not equal to or implied by some output of self.

        ``self`` does not imply ``other`` if any input of ``other``
        is not equal to or implied by some input of ``self``.

        ``self`` does not imply ``other`` if any despite of ``other``
        is not equal to or implied by some despite or input of ``self``.

        :returns:
            whether the assertion that ``self`` applies in some cases
            implies that the :class:`.Procedure` ``other`` applies
            in some cases.
        """

        if isinstance(other, self.__class__):
            yield from self._implies_procedure_if_present(other=other, context=context)

    def _explanations_same_meaning_as_procedure(
        self, other: Procedure, context: Explanation
    ):
        def same_outputs(context: Explanation):
            for explanation in self.outputs.explanations_same_meaning(
                other=other.outputs, context=context
            ):
                yield explanation

        def same_inputs(contexts: Iterable[Explanation]):
            for context in contexts:
                for explanation in self.inputs.explanations_same_meaning(
                    other=other.inputs, context=context
                ):
                    yield explanation

        def same_despite(contexts: Iterable[Explanation]):
            for context in contexts:
                for explanation in self.despite.explanations_same_meaning(
                    other=other.despite, context=context
                ):
                    yield explanation

        yield from same_despite(same_inputs(same_outputs(context)))

    def explanations_same_meaning(
        self,
        other: Comparable,
        context: Optional[Union[ContextRegister, Explanation]] = None,
    ) -> Iterator[Explanation]:
        """Yield contexts that could cause self to have the same meaning as other."""

        if not isinstance(context, Explanation):
            context = Explanation.from_context(context)

        if isinstance(other, self.__class__):
            yield from self._explanations_same_meaning_as_procedure(other, context)

    def means(
        self, other: Comparable, context: Optional[ContextRegister] = None
    ) -> bool:
        r"""
        Determine whether ``other`` has the same meaning as ``self``.

        :returns:
            whether the two :class:`.Procedure`\s have all the same
            :class:`.Factor`\s with the same context factors in the
            same roles.
        """
        return any(
            context is not None
            for context in self.explanations_same_meaning(other, context)
        )

    @new_context_helper
    def new_context(self, changes: ContextRegister) -> Procedure:
        r"""
        Create new :class:`Procedure`, replacing keys of ``changes`` with values.

        :param changes:
            a :class:`dict` of :class:`.Factor`\s to replace
            matched to the :class:`.Factor`\s that should replace them

        :returns:
            new :class:`.Procedure` object, replacing keys of
            ``changes`` with their values.
        """
        new_dict = self.__dict__.copy()
        for name in self.context_factor_names:
            new_dict[name] = tuple(
                factor.new_context(changes) for factor in new_dict[name]
            )
        new_dict.pop("context_factor_names")
        return self.__class__(**new_dict)

    def set_inputs(self, factors: Sequence[Factor]) -> None:
        self.inputs = FactorGroup(factors)

    def set_despite(self, factors: Sequence[Factor]) -> None:
        self.despite = FactorGroup(factors)

    def set_outputs(self, factors: Sequence[Factor]) -> None:
        self.outputs = FactorGroup(factors)

    def triggers_next_procedure(
        self,
        other: Procedure,
        context: Optional[Union[ContextRegister, Explanation]] = None,
    ) -> Iterator[Explanation]:
        r"""
        Test if :class:`.Factor`\s from firing ``self`` would trigger ``other``.

        .. Note::
            To be confident that ``self`` actually can trigger ``other``,
            we would have to assume that self or other has ``universal: True``
            since otherwise there could be a mismatch between what is provided
            and what is needed to trigger ``other``.

        :param other:
            another :class:`Procedure` to test to see whether it can
            be triggered by triggering ``self``

        :returns:
            whether the set of :class:`Factor`\s that exist after ``self``
            is fired could trigger ``other``
        """
        self_despite_or_input = FactorGroup((*self.despite, *self.inputs))
        self_output_or_input = FactorGroup((*self.outputs, *self.inputs))

        if not isinstance(context, Explanation):
            context = Explanation.from_context(context)

        for explanation_1 in self_output_or_input.explanations_implication(
            other.inputs, context=context
        ):
            yield from self_despite_or_input.explanations_implication(
                other.despite, context=explanation_1
            )

    def __or__(self, other: Comparable) -> Optional[Comparable]:
        return self.union(other)

    def union(
        self,
        other: Comparable,
        context: Optional[Union[ContextRegister, Explanation]] = None,
    ) -> Optional[Comparable]:
        if not isinstance(context, Explanation):
            context = Explanation.from_context(context)
        explanations = self.explanations_union(other, context)
        try:
            explanation = next(explanations)
        except StopIteration:
            return None
        return self._union_from_explanation(other, explanation)

    def valid_for_exclusive_tag(self) -> None:
        if len(self.outputs) != 1:
            raise ValueError(
                "The 'exclusive' attribute is not allowed for Holdings "
                + "with more than one 'output' Factor. If the set of Factors "
                + "in 'inputs' is really the only way to reach any of the "
                + "'outputs', consider making a separate 'exclusive' Rule "
                + "for each output."
            )
        if self.outputs[0].absent:
            raise ValueError(
                "The 'exclusive' attribute is not allowed for Holdings "
                + "with an 'absent' 'output' Factor. This would indicate "
                + "that the output can or must be present in every litigation "
                + "unless specified inputs are present, which is unlikely."
            )
        if not self.inputs:
            raise ValueError(
                "The 'exclusive' attribute is not allowed for Holdings "
                + "with no 'input' Factors, since the 'exclusive' attribute "
                + "asserts that the inputs are the only way to reach the output."
            )
        return None



"""
Converting simple structured data from XML or JSON into authorityspoke objects.

These functions will usually be called by functions from the io.loaders module
after they import some data from a file.
"""
from copy import deepcopy
from typing import Any, NamedTuple

from typing import Dict, List, Optional, Tuple, Type, Union

from anchorpoint.textselectors import TextQuoteSelector
from legislice.download import Client
from nettlesome.entities import Entity
from nettlesome.factors import Factor

from authorityspoke.decisions import Decision
from authorityspoke.evidence import Exhibit, Evidence
from authorityspoke.facts import Fact
from authorityspoke.holdings import Holding
from authorityspoke.io import schemas_yaml
from authorityspoke.opinions import AnchoredHoldings
from authorityspoke.pleadings import Allegation, Pleading
from authorityspoke.procedures import Procedure
from authorityspoke.rules import Rule
from authorityspoke.io import schemas_json
from authorityspoke.io.schemas_yaml import (
    RawEnactment,
    RawHolding,
    RawRule,
    RawPredicate,
    RawFactor,
    RawDecision,
    RawSelector,
)
from authorityspoke.io.name_index import index_names, Mentioned
from authorityspoke.io.enactment_index import collect_enactments

FACTOR_SUBCLASSES = {
    class_obj.__name__: class_obj
    for class_obj in (Allegation, Entity, Exhibit, Evidence, Fact, Pleading)
}


def read_fact(record: RawFactor) -> Fact:
    r"""
    Construct a :class:`Fact` after loading a dict from YAML.

    :param record:
        parameter values to pass to :class:`authorityspoke.schemas_yaml.FactSchema`\.

    :returns:
        a :class:`Fact`, with optional mentioned factors
    """
    record, mentioned = index_names(record)
    schema = schemas_yaml.FactSchema()
    schema.context["mentioned"] = mentioned
    return schema.load(record)


def read_factor(record: RawFactor, client: Optional[Client] = None, **kwargs) -> Factor:
    r"""
    Turn fields from YAML into a :class:`Factor` object.

    :param record:
        parameter values to pass to schema

    :param client:
        to look up any :class:`.Enactment` references

    """
    schema = schemas_yaml.FactorSchema(many=False)
    record, enactment_index = collect_enactments(record)
    if client:
        enactment_index = client.update_entries_in_enactment_index(enactment_index)
    schema.context["enactment_index"] = enactment_index
    record, schema.context["mentioned"] = index_names(record)

    return schema.load(record)


def read_factors(
    record: List[RawFactor], client: Optional[Client] = None, **kwargs
) -> Factor:
    r"""
    Turn fields from YAML into a :class:`Factor` object.

    :param record:
        parameter values to pass to schema

    :parame regime:
        to look up any :class:`.Enactment` references
    """
    schema = schemas_yaml.FactorSchema(many=True)
    record, enactment_index = collect_enactments(record)
    if client:
        enactment_index = client.update_entries_in_enactment_index(enactment_index)
    schema.context["enactment_index"] = enactment_index
    record, schema.context["mentioned"] = index_names(record)
    return schema.load(record)


def read_procedure(
    record: Dict, client: Optional[Client] = None, many=False
) -> Procedure:
    r"""
    Turn fields from YAML into a :class:`Procedure` object.

    :param record:
        parameter values to pass to schema

    :param many:
        whether to use the "many" form of the Marshmallow
        schema (whether there are multiple Procedures)

    :parame regime:
        to look up any :class:`.Enactment` references
    """
    schema = schemas_yaml.ProcedureSchema(many=many)
    record, enactment_index = collect_enactments(record)
    if client:
        enactment_index = client.update_entries_in_enactment_index(enactment_index)
    schema.context["enactment_index"] = enactment_index
    record, schema.context["mentioned"] = index_names(record)
    return schema.load(record)


def read_holding(record: RawHolding, client: Optional[Client] = None) -> Holding:
    r"""
    Create new :class:`Holding` object from simple datatypes from JSON input.

    Will yield multiple items if ``exclusive: True`` is present in ``record``.

    :param record:
        dict of values for constructing :class:`.Holding`

    :param client:
        Legislice client for downloading missing fields from `record`

    :param many:
        if True, record represents a list of :class:`Holding`\s rather than
        just one.

    :returns:
        New :class:`.Holding`, and an updated dictionary with mentioned
        :class:`.Factor`\s as keys and their :class:`.TextQuoteSelector`\s
        as values.
    """

    schema = schemas_yaml.HoldingSchema(many=False)

    record, enactment_index = collect_enactments(record)
    if client:
        enactment_index = client.update_entries_in_enactment_index(enactment_index)
    anchors, enactment_index = collect_anchors_from_index(enactment_index, "enactment")
    schema.context["enactment_index"] = enactment_index

    record, factor_index = index_names(record)
    anchors, factor_index = collect_anchors_from_index(factor_index, "name")
    schema.context["mentioned"] = factor_index

    return schema.load(deepcopy(record))


class HoldingsIndexed(NamedTuple):
    """Lists :class:`.Holding` objects with corresponding text selectors."""

    holdings: List[Holding]
    mentioned: Mentioned
    holding_anchors: List[List[TextQuoteSelector]]


def collect_anchors_from_index(object_index, field_name: str):
    anchors = []
    for key, value in object_index.items():
        if value.get("anchors"):
            anchored_object: Dict[str, Any] = {}
            anchored_object["anchors"] = value.pop("anchors")
            anchored_object[field_name] = value
            anchors.append(anchored_object)
    return anchors, object_index


def read_holdings_with_anchors(
    record: Dict[str, Union[List[RawHolding], List[RawSelector]]],
    client: Optional[Client] = None,
    many: bool = True,
) -> AnchoredHoldings:
    r"""
    Load a list of Holdings from JSON, with text links.

    :param record:
        a list of dicts representing holdings, in the JSON input format

    :param client:
        Legislice client for downloading missing fields from `record`

    :param many:
        a bool indicating whether to use the "many" form of the Marshmallow
        schema (whether there are multiple Holdings)

    :returns:
        a namedtuple listing :class:`.Holding` objects with
        a list matching :class:`.Holding`\s to selectors and
        an index matching :class:`.Factor`\s to selectors.
    """
    schema = schemas_yaml.AnchoredHoldingsSchema()

    record, enactment_index = collect_enactments(record)
    if client:
        enactment_index = client.update_entries_in_enactment_index(enactment_index)

    # move anchors out of enactments
    anchors, enactment_index = collect_anchors_from_index(enactment_index, "enactment")
    if not record.get("enactment_anchors"):
        record["enactment_anchors"] = []
    record["enactment_anchors"] = record["enactment_anchors"] + anchors
    schema.context["enactment_index"] = enactment_index

    record["holdings"], factor_index = index_names(record["holdings"])

    # move anchors out of factors
    anchors, factor_index = collect_anchors_from_index(factor_index, "name")
    if not record.get("factor_anchors"):
        record["factor_anchors"] = []
    record["factor_anchors"] = record["factor_anchors"] + anchors
    schema.context["mentioned"] = factor_index

    holdings, holding_anchors, named_anchors, enactment_anchors = schema.load(
        deepcopy(record)
    )
    return AnchoredHoldings(holdings, holding_anchors, named_anchors, enactment_anchors)


def read_holdings(
    record: List[RawHolding], client: Optional[Client] = None
) -> List[Holding]:
    r"""
    Load a list of :class:`Holdings`\s from JSON.

    :param record:
        a list of dicts representing holdings, in the JSON input format

    :parame regime:
        A collection of :class:`.Jurisdiction`\s and the :class:`.Code`\s
        that have been enacted in each. Used for constructing
        :class:`.Enactment`\s referenced by :class:`.Holding`\s.

    :returns:
        a list of :class:`.Holding` objects
    """
    schema = schemas_yaml.HoldingSchema(many=True)

    record, enactment_index = collect_enactments(record)
    if client:
        enactment_index = client.update_entries_in_enactment_index(enactment_index)
    anchors, enactment_index = collect_anchors_from_index(enactment_index, "enactment")
    schema.context["enactment_index"] = enactment_index

    record, factor_index = index_names(record)
    anchors, factor_index = collect_anchors_from_index(factor_index, "name")
    schema.context["mentioned"] = factor_index

    return schema.load(deepcopy(record))


def read_decision(decision_dict: RawDecision) -> Decision:
    r"""
    Create and return one or more :class:`.Opinion` objects from a dict API response.

    Relies on the JSON format from the `Caselaw Access Project
    API <https://api.case.law/v1/cases/>`_.

    This function is a more convenient way to call read_opinions with an entire
    case from the CAP API as a single parameter.

    :param decision_dict:
        A dict created from a Caselaw Access Project API response.
    """
    schema = schemas_json.DecisionSchema()
    return schema.load(decision_dict)


def read_rules_with_index(
    record: List[RawRule], client: Optional[Client] = None, many: bool = True
) -> Tuple[List[Rule], Mentioned]:
    r"""Make :class:`Rule` and "mentioned" index from dict of fields and :class:`.Regime`\."""

    schema = schemas_yaml.RuleSchema(many=many)
    record, enactment_index = collect_enactments(record)

    if client:
        enactment_index = client.update_entries_in_enactment_index(enactment_index)
    record, mentioned = index_names(record)
    schema.context["enactment_index"] = enactment_index

    schema.context["mentioned"] = mentioned

    rules = schema.load(deepcopy(record))
    return rules, mentioned


def read_rule(record: Dict, client: Optional[Client] = None) -> Rule:
    r"""
    Make :class:`Rule` from a :class:`dict` of fields and a :class:`.Regime`\.

    :param record:

    :param client:
        Legislice client for downloading missing fields from `record`

    :returns:
        iterator yielding :class:`Rule`\s with the items
        from ``mentioned_entities`` as ``terms``
    """
    schema = schemas_yaml.RuleSchema()
    record, enactment_index = collect_enactments(record)
    record, schema.context["mentioned"] = index_names(record)

    if client:
        enactment_index = client.update_entries_in_enactment_index(enactment_index)
    schema.context["enactment_index"] = enactment_index
    return schema.load(record)


def read_rules(record: List[Dict], client: Optional[Client] = None) -> List[Rule]:
    r"""
    Make :class:`Rule`\s from a list of fields and a :class:`.Regime`\.

    :param record:

    :param client:
        Legislice client for downloading missing fields from `record`

    :returns:
        iterator yielding :class:`Rule`\s with the items
        from ``mentioned_entities`` as ``terms``
    """

    schema = schemas_yaml.RuleSchema(many=True)
    record, enactment_index = collect_enactments(record)
    record, schema.context["mentioned"] = index_names(record)

    if client:
        enactment_index = client.update_entries_in_enactment_index(enactment_index)
    schema.context["enactment_index"] = enactment_index
    return schema.load(record)



"""
Marshmallow schemas for loading AuthoritySpoke objects from YAML.

Intended for use with human-generated files.
Some shortened syntax is allowed.
"""

from copy import deepcopy
from datetime import date
from typing import Any, Dict, List, NamedTuple, Optional, Sequence, Tuple, Type, Union

from marshmallow import Schema, fields, validate, EXCLUDE
from marshmallow import pre_load, post_load
from marshmallow import ValidationError

from anchorpoint.textselectors import TextQuoteSelector, TextPositionSelector
from anchorpoint.schemas import SelectorSchema
from legislice import Enactment
from legislice.schemas import enactment_needs_api_update
from legislice.yaml_schemas import ExpandableEnactmentSchema as LegisliceSchema

from nettlesome.entities import Entity
from nettlesome.factors import Factor
from nettlesome.predicates import Predicate
from nettlesome.quantities import Comparison, QuantityRange, Quantity
from nettlesome.schemas import dump_quantity

from authorityspoke.decisions import CAPCitation, Decision
from authorityspoke.evidence import Exhibit, Evidence

from authorityspoke.facts import Fact
from authorityspoke.holdings import Holding
from authorityspoke.io.enactment_index import EnactmentIndex
from authorityspoke.io.name_index import Mentioned
from authorityspoke.io.name_index import RawFactor, RawPredicate
from authorityspoke.io.nesting import nest_fields
from authorityspoke.io import text_expansion

from authorityspoke.opinions import AnchoredHoldings
from authorityspoke.pleadings import Pleading, Allegation

from authorityspoke.procedures import Procedure
from authorityspoke.rules import Rule

from authorityspoke.utils.marshmallow_oneofschema.one_of_schema import OneOfSchema

RawSelector = Union[str, Dict[str, str]]
RawEnactment = Dict[str, Union[str, List[RawSelector]]]
RawProcedure = Dict[str, Sequence[RawFactor]]
RawRule = Dict[str, Union[RawProcedure, Sequence[RawEnactment], str, bool]]
RawHolding = Dict[str, Union[RawRule, str, bool]]


class ExpandableSchema(Schema):
    """Base schema for classes that can be cross-referenced by name in input JSON."""

    def get_from_mentioned(self, data, **kwargs):
        """Replace data to load with any object with same name in "mentioned"."""
        if isinstance(data, str):
            mentioned = self.context.get("mentioned") or Mentioned()
            try:
                return deepcopy(mentioned.get_by_name(data))
            except ValueError:
                print(data)
        return data

    def consume_type_field(self, data, **kwargs):
        """Verify that type field is correct and then get rid of it."""
        if data.get("type"):
            data.pop("type")
        return data

    def wrap_single_element_in_list(self, data: Dict, many_element: str):
        """Make a specified field a list if it isn't already a list."""
        if data.get(many_element) is not None and not isinstance(
            data[many_element], list
        ):
            data[many_element] = [data[many_element]]
        return data

    @pre_load
    def format_data_to_load(self, data: RawFactor, **kwargs) -> RawFactor:
        """Expand data if it was just a name reference in the JSON input."""
        data = self.get_from_mentioned(data)
        data = self.consume_type_field(data)
        return data

    @post_load
    def make_object(self, data, **kwargs):
        """Make AuthoritySpoke object out of whatever data has been loaded."""
        if data.get("expression") is not None:
            return Comparison(**data)
        data.pop("expression", None)
        data.pop("sign", None)
        return self.__model__(**data)


RawOpinion = Dict[str, str]
RawCAPCitation = Dict[str, str]
RawDecision = Dict[str, Union[str, int, Sequence[RawOpinion], Sequence[RawCAPCitation]]]


class EnactmentSchema(LegisliceSchema):
    """
    Schema to load Enactments from JSON created from user-created YAML.

        >>> enactment_data = {
        ...     "name": "original works",
        ...     "node": "/us/usc/t17/s102",
        ...     "start_date": "1990-12-01",
        ...     "content": (
        ...         "Copyright protection subsists, in accordance with this title, "
        ...         "in original works of authorship fixed in any tangible medium of expression, "
        ...         "now known or later developed."),
        ...     "selection": [{"suffix": ", in accordance"}, "title,|in original works of authorship|fixed"]}
        >>> schema = EnactmentSchema()
        >>> enactment_index = EnactmentIndex()
        >>> enactment_index.insert_by_name(enactment_data)
        >>> schema.context["enactment_index"] = enactment_index
        >>> enactment = schema.load("original works")
        >>> enactment.selected_text()
        'Copyright protection subsists…in original works of authorship…'
    """

    def get_indexed_enactment(self, data, **kwargs):
        """Replace data to load with any object with same name in "enactment_index"."""
        if not isinstance(data, str):
            return data

        mentioned = self.context.get("enactment_index") or EnactmentIndex()
        return deepcopy(mentioned.get_by_name(data))

    @pre_load
    def format_data_to_load(self, data, **kwargs):
        """Prepare Enactment to load."""
        data = self.get_indexed_enactment(data)
        return super().format_data_to_load(data)


class PredicateSchema(ExpandableSchema):
    """
    Schema for statements, separate from any claim about their truth or who asserts them.

        >>> data = {
        ...     "content": "the distance that $officer pursued $suspect was >= 5 miles"
        ... }
        >>> schema = PredicateSchema()
        >>> predicate = schema.load(data)
        >>> predicate.quantity
        <Quantity(5, 'mile')>
    """

    __model__ = Predicate
    content = fields.Method(
        serialize="get_content_with_placeholders", deserialize="get_content_field"
    )
    truth = fields.Bool(missing=True)
    sign = fields.Str(
        missing=None,
        validate=validate.OneOf([""] + list(QuantityRange.opposite_comparisons.keys())),
    )
    expression = fields.Function(
        dump_quantity, deserialize=Comparison.read_quantity, missing=None
    )

    def get_content_with_placeholders(self, obj) -> str:
        return obj.template.template

    def get_content_field(self, data) -> str:
        return str(data)

    def split_quantity_from_content(
        self, content: str
    ) -> Tuple[str, Optional[str], Optional[Union[Quantity, int, float]]]:
        """Find any reference to a quantity in the content text."""
        for comparison in {
            **QuantityRange.opposite_comparisons,
            **QuantityRange.normalized_comparisons,
        }:
            if comparison in content:
                content, quantity_text = content.split(comparison)
                return content.rstrip(), comparison, quantity_text.lstrip()
        return content, "", None

    def normalize_comparison(self, data: RawPredicate, **kwargs) -> RawPredicate:
        """Reduce the number of possible symbols to represent comparisons."""
        if data.get("expression") and not data.get("sign"):
            data["sign"] = "="

        if data.get("sign") in QuantityRange.normalized_comparisons:
            data["sign"] = QuantityRange.normalized_comparisons[data["sign"]]
        return data

    @pre_load
    def format_data_to_load(self, data: RawPredicate, **kwargs) -> RawPredicate:
        """Expand any reference to a quantity in the content text."""
        if not data.get("expression"):
            (
                data["content"],
                data["sign"],
                data["expression"],
            ) = self.split_quantity_from_content(data["content"])
        data = self.normalize_comparison(data)
        return data


class EntitySchema(ExpandableSchema):
    """
    Schema for Entities, which shouldn't be at the top level of a FactorGroup.

        >>> entity = {"name": "Lotus Development Corporation", "generic": False}
        >>> schema = EntitySchema()
        >>> factor_index = Mentioned()
        >>> factor_index.insert_by_name(entity)
        >>> schema.context["mentioned"] = factor_index
        >>> loaded = schema.load("Lotus Development Corporation")
        >>> loaded.generic
        False

    """

    __model__: Type = Entity
    name = fields.Str(missing=None)
    generic = fields.Bool(missing=True)
    plural = fields.Bool()


class FactSchema(ExpandableSchema):
    """
    Schema for Facts, which may contain arbitrary levels of nesting.

        >>> data = {"content": "the distance that {Officer Lin} pursued {Al} was >= 5 miles"}
        >>> schema = FactSchema()
        >>> fact = schema.load(data)
        >>> print(fact)
        the fact that the distance that {Officer Lin} pursued {Al} was at least 5 mile

    """

    __model__: Type = Fact
    predicate = fields.Nested(PredicateSchema)
    terms = fields.Nested(lambda: FactorSchema(many=True))
    standard_of_proof = fields.Str(missing=None)
    name = fields.Str(missing=None)
    absent = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)

    def get_references_from_mentioned(
        self,
        content: str,
        terms: Optional[List[Dict]] = None,
        placeholder: str = "{}",
    ) -> Tuple[str, List[Dict]]:
        r"""
        Retrieve known context :class:`.Factor`\s for new :class:`.Fact`\.

        :param content:
            the content for the :class:`.Fact`\'s :class:`.Predicate`
        :param mentioned:
            list of :class:`.Factor`\s with names that could be
            referenced in content
        :param placeholder:
            a string to replace the names of
            referenced :class:`.Factor`\s in content
        :returns:
            the content string with any referenced :class:`.Factor`\s
            replaced by placeholder, and a list of referenced
            :class:`.Factor`\s in the order they appeared in content.
        """
        mentioned = self.context.get("mentioned") or Mentioned({})
        terms = terms or []
        for factor_name in mentioned.keys():
            if factor_name in content and factor_name != content:
                obj = mentioned.get_by_name(factor_name)
                content, terms = text_expansion.add_found_context(
                    content,
                    terms,
                    factor=deepcopy(obj),
                )
        return content, terms

    @pre_load
    def format_data_to_load(self, data: RawFactor, **kwargs) -> RawFactor:
        r"""
        Prepare :class:`.RawFact` to load, replacing name references with full objects.

        Unlike the :func:`.name_index.collect_mentioned` function, this function can't add
        any entries to the "mentioned" name index (due to limitations in the Marshmallow
        serialization library). That means all shorthand references to factors
        need to have been expanded before using the schema to load new objects.

        :param data:
            a dict representing a :class:`.Fact`

        :returns:
            a normalized dict representing a :class:`.Fact`\s with name references
            expanded
        """
        data = self.get_from_mentioned(data)
        to_nest = [
            "content",
            "truth",
            "sign",
            "expression",
        ]
        data = nest_fields(data, nest="predicate", eggs=to_nest)
        data = self.wrap_single_element_in_list(data, "terms")
        (
            data["predicate"]["content"],
            data["terms"],
        ) = self.get_references_from_mentioned(
            data["predicate"]["content"], data.get("terms")
        )
        data = self.consume_type_field(data)
        return data

    @post_load
    def make_object(self, data: RawFactor, **kwargs) -> Fact:
        """Make Fact."""
        return self.__model__(**data)


class ExhibitSchema(ExpandableSchema):
    """
    Schema for an object that may embody a statement.

        >>> fact_data = {
        ...     "content": "the distance that $officer pursued $suspect was >= 5 miles",
        ...     "terms": [
        ...         {"type": "Entity", "name": "Officer Lin"},
        ...         {"type": "Entity", "name": "Al"},
        ...     ],
        ... }
        >>> exhibit_data = {
        ...     "form": "testimony",
        ...     "statement": fact_data,
        ...     "statement_attribution": {"name": "Officer Lin"},
        ... }
        >>> schema = ExhibitSchema()
        >>> exhibit = schema.load(exhibit_data)
        >>> print(exhibit)
        the testimony attributed to <Officer Lin>, asserting the fact that the distance that <Officer Lin> pursued <Al> was at least 5 mile,

    """

    __model__: Type = Exhibit
    form = fields.Str(missing=None)
    statement = fields.Nested(FactSchema, missing=None)
    statement_attribution = fields.Nested(EntitySchema, missing=None)
    name = fields.Str(missing=None)
    absent = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)

    @post_load
    def make_object(self, data: RawFactor, **kwargs) -> Exhibit:
        """Make Exhibit."""
        return self.__model__(**data)


class PleadingSchema(ExpandableSchema):
    """Schema for a document to link Allegations to."""

    __model__: Type = Pleading
    filer = fields.Nested(EntitySchema, missing=None)
    name = fields.Str(missing=None)
    absent = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)


class AllegationSchema(ExpandableSchema):
    """Schema for an Allegation of a Fact."""

    __model__: Type = Allegation
    pleading = fields.Nested(PleadingSchema, missing=None)
    statement = fields.Nested(FactSchema, missing=None)
    name = fields.Str(missing=None)
    absent = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)


class EvidenceSchema(ExpandableSchema):
    """Schema for an Exhibit and a reference to the Fact it would support."""

    __model__: Type = Evidence
    exhibit = fields.Nested(ExhibitSchema, missing=None)
    to_effect = fields.Nested(FactSchema, missing=None)
    name = fields.Str(missing=None)
    absent = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)


class FactorSchema(OneOfSchema, ExpandableSchema):
    """
    Schema that directs data to "one of" the other schemas.

        >>> factor_data = [{"type": "Entity", "name": "Al"}, {"type": "Entity", "name": "Bob"}]
        >>> schema = FactorSchema(many=True)
        >>> factors = schema.load(factor_data)
        >>> factors[1].name
        'Bob'
    """

    __model__: Type = Factor
    type_schemas = {
        "allegation": AllegationSchema,
        "Allegation": AllegationSchema,
        "entity": EntitySchema,
        "Entity": EntitySchema,
        "evidence": EvidenceSchema,
        "Evidence": EvidenceSchema,
        "exhibit": ExhibitSchema,
        "Exhibit": ExhibitSchema,
        "fact": FactSchema,
        "Fact": FactSchema,
        "pleading": PleadingSchema,
        "Pleading": PleadingSchema,
    }

    def pre_load(self, data: RawFactor, **kwargs) -> RawFactor:
        data = self.get_from_mentioned(data)
        return data

    def get_obj_type(self, obj) -> str:
        """Return name of object schema."""
        return obj.__class__.__name__


class ProcedureSchema(ExpandableSchema):
    """
    Schema for Procedure; does not require separate TermSequence schema.

    "FactorSchema, many=True" is an equivalent of a TermSequence.
    """

    __model__: Type = Procedure
    inputs = fields.Nested(FactorSchema, many=True)
    despite = fields.Nested(FactorSchema, many=True)
    outputs = fields.Nested(FactorSchema, many=True)

    @pre_load
    def format_data_to_load(self, data, **kwargs):
        """Handle omission of brackets around single item."""
        for field_name in ("inputs", "despite", "outputs"):
            data = self.wrap_single_element_in_list(data, field_name)
        return data


class RuleSchema(ExpandableSchema):
    """Schema for Holding; can also hold Procedure fields."""

    __model__: Type = Rule
    procedure = fields.Nested(ProcedureSchema)
    enactments = fields.Nested(EnactmentSchema, many=True)
    enactments_despite = fields.Nested(EnactmentSchema, many=True)
    mandatory = fields.Bool(missing=False)
    universal = fields.Bool(missing=False)
    name = fields.Str(missing=None)
    generic = fields.Bool(missing=False)

    @pre_load
    def format_data_to_load(self, data: Dict, **kwargs) -> RawRule:
        """Prepare to load Rule."""
        data = self.wrap_single_element_in_list(data, "enactments")
        data = self.wrap_single_element_in_list(data, "enactments_despite")
        procedure_fields = ("inputs", "despite", "outputs")
        data["procedure"] = data.get("procedure") or {}
        for field in procedure_fields:
            if field in data:
                data["procedure"][field] = data.pop(field)
        return data

    @post_load
    def make_object(self, data, **kwargs):
        return self.__model__(**data)


class HoldingSchema(ExpandableSchema):
    """Schema for Holding; can also hold Rule and Procedure fields."""

    __model__: Type = Holding
    rule = fields.Nested(RuleSchema)
    rule_valid = fields.Bool(missing=True)
    decided = fields.Bool(missing=True)
    exclusive = fields.Bool(missing=False)
    generic = fields.Bool(missing=False)
    anchors = fields.Nested(SelectorSchema, many=True)

    def nest_fields_inside_rule(self, data: Dict) -> RawHolding:
        """Nest fields inside "rule" and "procedure", if not already nested."""
        data["rule"]["procedure"] = data["rule"].get("procedure") or {}
        procedure_fields = ("inputs", "despite", "outputs")
        for field in procedure_fields:
            if field in data:
                data["rule"]["procedure"][field] = data.pop(field)
            if field in data["rule"]:
                data["rule"]["procedure"][field] = data["rule"].pop(field)

        return data

    @pre_load
    def format_data_to_load(self, data: RawHolding, **kwargs) -> RawHolding:
        """Prepare to load Holding."""
        data = self.get_from_mentioned(data)

        data["rule"] = data.get("rule") or {}
        to_nest = [
            "procedure",
            "enactments",
            "enactments_despite",
            "mandatory",
            "universal",
        ]
        data = nest_fields(data, nest="rule", eggs=to_nest)
        data = self.nest_fields_inside_rule(data)

        return data


class NamedAnchors(NamedTuple):
    name: Factor
    anchors: List[TextQuoteSelector]


class NamedAnchorsSchema(ExpandableSchema):
    __model__ = NamedAnchors

    name = fields.Nested(FactorSchema)
    anchors = fields.Nested(SelectorSchema, many=True)

    @pre_load
    def format_data_to_load(self, data, **kwargs):
        data = self.wrap_single_element_in_list(data, "anchors")
        return data


class AnchoredEnactments(NamedTuple):
    enactment: Enactment
    anchors: List[TextQuoteSelector]


class AnchoredEnactmentsSchema(ExpandableSchema):
    __model__ = AnchoredEnactments

    enactment = fields.Nested(EnactmentSchema)
    anchors = fields.Nested(SelectorSchema, many=True)

    @pre_load
    def format_data_to_load(self, data, **kwargs):
        data = self.wrap_single_element_in_list(data, "anchors")
        return data


class AnchoredHoldingsSchema(ExpandableSchema):
    __model__ = AnchoredHoldings

    holdings = fields.Nested(HoldingSchema, many=True)
    factor_anchors = fields.Nested(NamedAnchorsSchema, many=True)
    enactment_anchors = fields.Nested(AnchoredEnactmentsSchema, many=True)

    @pre_load
    def format_data_to_load(self, data: RawFactor, **kwargs) -> RawFactor:
        """Expand data if it was just a name reference in the JSON input."""

        data["holdings"] = self.get_from_mentioned(data["holdings"])
        return data

    @post_load
    def make_object(self, data, **kwargs) -> AnchoredHoldings:
        """Make AuthoritySpoke object out of whatever data has been loaded."""
        text_links = {}
        enactment_links = {}

        if data.get("factor_anchors"):
            for linked in data["factor_anchors"]:
                text_links[linked.name.key] = linked.anchors

        if data.get("enactment_anchors"):
            for linked in data["enactment_anchors"]:
                if not linked.enactment.selected_text():
                    linked.enactment.select_all()
                enactment_links[str(linked.enactment)] = linked.anchors

        holding_anchors = [holding.anchors for holding in data["holdings"]]
        return AnchoredHoldings(
            data["holdings"],
            holding_anchors,
            named_anchors=text_links,
            enactment_anchors=enactment_links,
        )



from datetime import date
from typing import Type
import pytest
import sympy
from sympy import Interval, oo


from nettlesome.entities import Entity
from nettlesome.predicates import Predicate
from nettlesome.quantities import Comparison, Q_, Quantity


class TestComparisons:
    def test_comparison_with_wrong_comparison_symbol(self):
        with pytest.raises(ValueError):
            _ = Comparison(
                "the height of {} was {}",
                sign=">>",
                expression=Q_("160 centimeters"),
            )

    def test_comparison_interval(self):
        comparison = Comparison(
            "the distance between $place1 and $place2 was",
            sign=">",
            expression=Q_("20 miles"),
        )
        assert comparison.interval == Interval(20, oo, left_open=True)

    def test_comparison_not_equal(self):
        comparison = Comparison(
            "the distance between $place1 and $place2 was",
            sign="!=",
            expression=Q_("20 miles"),
        )
        assert comparison.interval == sympy.Union(
            Interval(0, 20, right_open=True), Interval(20, oo, left_open=True)
        )


class TestPredicates:
    def test_no_sign_allowed_for_predicate(self):
        with pytest.raises(TypeError):
            Predicate(
                "the date when $work was created was",
                sign=">=",
                expression=date(1978, 1, 1),
            )

    def test_term_positions(self):
        predicate = Predicate(
            content="$organizer1 and $organizer2 planned for $player1 to play $game with $player2."
        )
        assert predicate.term_positions() == {
            "organizer1": {0, 1},
            "organizer2": {0, 1},
            "player1": {2, 4},
            "game": {3},
            "player2": {2, 4},
        }

    def test_term_positions_with_repetition(self):
        predicate = Predicate(
            content="$organizer1 and $organizer2 planned for $organizer1 to play $game with $organizer2."
        )
        assert predicate.term_positions() == {
            "organizer1": {0, 1},
            "organizer2": {0, 1},
            "game": {2},
        }

    def test_term_permutations(self):
        predicate = Predicate(
            content="$organizer1 and $organizer2 planned for $player1 to play $game with $player2."
        )
        assert predicate.term_index_permutations() == [
            (0, 1, 2, 3, 4),
            (0, 1, 4, 3, 2),
            (1, 0, 2, 3, 4),
            (1, 0, 4, 3, 2),
        ]

    def test_term_permutations_with_repetition(self):
        predicate = Predicate(
            content="$organizer1 and $organizer2 planned for $organizer1 to play $game with $organizer2."
        )
        assert predicate.term_index_permutations() == [
            (0, 1, 2),
            (1, 0, 2),
        ]

    def test_convert_false_statement_about_quantity_to_obverse(self, make_predicate):
        assert make_predicate["p7_obverse"].truth is True
        assert make_predicate["p7_obverse"].quantity == Q_(35, "foot")
        assert make_predicate["p7"].truth is True
        assert make_predicate["p7"].sign == "<="
        assert 'sign="<="' in repr(make_predicate["p7"])
        assert make_predicate["p7_obverse"].sign == "<="

    def test_quantity_type(self, make_predicate):
        assert isinstance(make_predicate["p7"].quantity, Quantity)

    def test_string_for_date_as_expression(self):
        copyright_date_range = Comparison(
            "the date when $work was created was",
            sign=">=",
            expression=date(1978, 1, 1),
        )
        assert str(copyright_date_range).endswith("1978-01-01")

    def test_quantity_string(self, make_predicate):
        assert str(make_predicate["p7"].quantity) == "35 foot"

    def test_predicate_content_comparison(self, make_predicate):
        assert make_predicate["p8_exact"].content == make_predicate["p7"].content

    def test_expression_comparison(self, make_predicate):
        assert str(make_predicate["p7"].quantity_range) == "no more than 35 foot"
        assert str(make_predicate["p9"].quantity_range) == "no more than 5 foot"

    def test_predicate_has_no_expression_comparison(self, make_predicate):
        with pytest.raises(AttributeError):
            make_predicate["p1"].expression_comparison() == ""

    def test_context_slots(self, make_predicate):
        assert len(make_predicate["p7"]) == 2

    def test_str_for_predicate_with_number_quantity(self, make_predicate):
        assert "distance between $place1 and $place2 was at least 20" in str(
            make_predicate["p8_int"]
        )
        assert "distance between $place1 and $place2 was at least 20.0" in str(
            make_predicate["p8_float"]
        )
        assert "distance between $place1 and $place2 was at least 20 foot" in str(
            make_predicate["p8"]
        )

    def test_template_singular_by_default(self):
        predicate = Predicate("$people were in $city")
        assert str(predicate.template) == 'StatementTemplate("$people was in $city")'

    @pytest.mark.parametrize(
        "context, expected",
        [
            (
                [Entity(name="the book", plural=False)],
                "<the book> was names, towns,",
            ),
            (
                [Entity(name="the book's listings", plural=True)],
                "<the book's listings> were names, towns,",
            ),
        ],
    )
    def test_make_str_plural(self, context, expected):
        phrase = (
            "$thing were names, towns, and telephone numbers of telephone subscribers"
        )
        predicate = Predicate(phrase)
        with_context = predicate._content_with_terms(context)
        assert with_context.startswith(expected)

    def test_str_not_equal(self, make_predicate):
        assert (
            "the distance between $place1 and $place2 was not equal to 35 foot"
            in str(make_predicate["p7_not_equal"])
        )

    def test_negated_method(self, make_predicate):
        assert make_predicate["p7"].negated().means(make_predicate["p7_opposite"])
        assert make_predicate["p3"].negated().means(make_predicate["p3_false"])


class TestSameMeaning:
    def test_predicate_equality(self, make_predicate):
        assert make_predicate["p1"].means(make_predicate["p1_again"])

    def test_predicate_inequality(self, make_predicate, watt_factor):
        assert not make_predicate["p2"].means(make_predicate["p2_reflexive"])

    def test_error_predicate_means_fact(self, make_predicate, watt_factor):
        with pytest.raises(TypeError):
            make_predicate["p2"].means(watt_factor["f2"])

    def test_obverse_predicates_equal(self, make_predicate):
        assert make_predicate["p7"].means(make_predicate["p7_obverse"])

    def test_equal_float_and_int(self, make_predicate):
        """
        These now evaluate equal even though their equal quantities are different types
        """
        assert make_predicate["p8_int"].means(make_predicate["p8_float"])

    def test_same_meaning_float_and_int(self, make_predicate):
        """
        The Predicate means method considers equal quantities of different types to have the same meaning.
        """
        assert make_predicate["p8_int"].means(make_predicate["p8_float"])

    def test_no_equality_with_inconsistent_dimensionality(self, make_predicate):
        assert not make_predicate["p9"].means(make_predicate["p9_acres"])

    def test_different_truth_value_prevents_equality(self, make_predicate):
        assert not make_predicate["p_murder"].means(make_predicate["p_murder_whether"])
        assert not make_predicate["p_murder_false"].means(
            make_predicate["p_murder_whether"]
        )
        assert not make_predicate["p_murder_false"].means(make_predicate["p_murder"])

    def test_predicate_does_not_mean_fact(self, make_predicate, watt_factor):
        with pytest.raises(TypeError):
            make_predicate["p8"].means(watt_factor["f8"])

    def test_term_placeholders_do_not_change_result(self):
        left = Predicate(
            content="$organizer1 and $organizer2 planned for $player1 to play $game with $player2."
        )
        right = Predicate(
            content="$promoter1 and $promoter2 planned for $player1 to play $chess with $player2."
        )
        assert left.means(right)

    def test_term_positions_change_result(self):
        left = Predicate(
            content="$organizer1 and $organizer2 planned for $player1 to play $game with $player2."
        )
        right = Predicate(
            content="$organizer1 and $organizer2 planned for $organizer1 to play $game with $organizer2."
        )
        assert not left.means(right)


class TestImplication:
    def test_greater_than_because_of_quantity(self, make_predicate):
        assert make_predicate["p8_meters"] > make_predicate["p8"]
        assert make_predicate["p8_meters"] != make_predicate["p8"]

    def test_greater_float_and_int(self, make_predicate):
        assert make_predicate["p8_higher_int"] > make_predicate["p8_float"]
        assert make_predicate["p8_int"] < make_predicate["p8_higher_int"]

    def test_any_truth_value_implies_none(self, make_predicate):
        assert make_predicate["p_murder"] > make_predicate["p_murder_whether"]
        assert make_predicate["p_murder_false"] > make_predicate["p_murder_whether"]

    def test_no_implication_by_exact_quantity(self, make_predicate):
        assert not make_predicate["p_quantity=3"] > make_predicate["p_quantity>5"]

    def test_no_implication_of_exact_quantity(self, make_predicate):
        assert not make_predicate["p_quantity>5"] > make_predicate["p_quantity=3"]

    def test_no_implication_by_greater_or_equal_quantity(self, make_predicate):
        assert not make_predicate["p_quantity>=4"] > make_predicate["p_quantity>5"]

    def test_no_implication_of_greater_or_equal_quantity(self):
        less = Comparison(content="The number of mice was", sign=">", expression=4)
        more = Comparison(content="The number of mice was", sign=">=", expression=5)
        assert not less.implies(more)

    def test_no_contradiction_inconsistent_dimensions(self):
        equal = Comparison(
            "${defendant}'s sentence was", sign="=", expression="8 years"
        )
        less = Comparison(
            "${defendant}'s sentence was", sign="<=", expression="10 parsecs"
        )
        assert not equal.contradicts(less)
        assert not equal.implies(less)

    def test_equal_implies_greater_or_equal(self, make_predicate):
        assert make_predicate["p9_exact"] > make_predicate["p9"]

    def test_implication_with_not_equal(self, make_predicate):
        assert make_predicate["p7_opposite"] > make_predicate["p7_not_equal"]

    def test_no_implication_with_inconsistent_dimensionality(self, make_predicate):
        assert not make_predicate["p9"] >= make_predicate["p9_acres"]
        assert not make_predicate["p9"] <= make_predicate["p9_acres"]

    def test_implication_with_no_truth_value(self, make_predicate):
        assert not make_predicate["p2_no_truth"] > make_predicate["p2"]
        assert make_predicate["p2"] > make_predicate["p2_no_truth"]

    def test_predicate_cannot_imply_factor(self, make_predicate, watt_factor):
        assert not make_predicate["p7_true"] > watt_factor["f7"]

    def test_implication_due_to_dates(self):
        copyright_date_range = Comparison(
            "the date when $work was created was",
            sign=">=",
            expression=date(1978, 1, 1),
        )
        copyright_date_specific = Comparison(
            "the date when $work was created was",
            sign="=",
            expression=date(1980, 6, 20),
        )
        assert copyright_date_specific.implies(copyright_date_range)


class TestContradiction:
    def test_predicate_no_contradictions(self, make_predicate):
        assert not make_predicate["p7"].contradicts(make_predicate["p7_true"])
        assert not make_predicate["p1"].contradicts(make_predicate["p1_again"])
        assert not make_predicate["p3"].contradicts(make_predicate["p7"])

    def test_contradiction_by_exact(self, make_predicate):
        assert make_predicate["p8_exact"].contradicts(make_predicate["p8_less"])

    def test_contradiction_of_exact(self, make_predicate):
        assert make_predicate["p8_less"].contradicts(make_predicate["p8_exact"])

    def test_contradiction_by_equal_quantity(self, make_predicate):
        assert make_predicate["p_quantity=3"].contradicts(
            make_predicate["p_quantity>5"]
        )

    def test_contradiction_of_equal_quantity(self, make_predicate):
        assert make_predicate["p_quantity>5"].contradicts(
            make_predicate["p_quantity=3"]
        )

    def test_no_contradiction_by_greater_or_equal_quantity(self, make_predicate):
        assert not make_predicate["p_quantity>=4"].contradicts(
            make_predicate["p_quantity>5"]
        )

    def test_no_contradiction_of_greater_or_equal_quantity(self, make_predicate):
        assert not make_predicate["p_quantity>5"].contradicts(
            make_predicate["p_quantity>=4"]
        )

    def test_error_predicate_contradict_factor(self, make_predicate, watt_factor):
        with pytest.raises(TypeError):
            make_predicate["p7_true"].contradicts(watt_factor["f7"])

    def test_no_contradiction_with_no_truth_value(self, make_predicate):
        assert not make_predicate["p2_no_truth"].contradicts(make_predicate["p2"])
        assert not make_predicate["p2"].contradicts(make_predicate["p2_no_truth"])

    def test_no_contradiction_with_inconsistent_dimensionality(self, make_predicate):
        assert not make_predicate["p9"].contradicts(make_predicate["p9_acres"])
        assert not make_predicate["p9_acres"].contradicts(make_predicate["p9"])

    def test_contradiction_with_quantity(self, make_predicate):
        assert make_predicate["p8_less"].contradicts(make_predicate["p8_meters"])

    def test_contradictory_date_ranges(self):
        later = Comparison(
            "the date $dentist became a licensed dentist was",
            sign=">",
            expression=date(2010, 1, 1),
        )
        earlier = Comparison(
            "the date $dentist became a licensed dentist was",
            sign="<",
            expression=date(1990, 1, 1),
        )
        assert later.contradicts(earlier)
        assert earlier.contradicts(later)

    def test_no_contradiction_without_truth_value(self):
        later = Comparison(
            "the date $dentist became a licensed dentist was",
            sign=">",
            expression=date(2010, 1, 1),
            truth=None,
        )
        earlier = Comparison(
            "the date $dentist became a licensed dentist was",
            sign="<",
            expression=date(1990, 1, 1),
        )
        assert not later.contradicts(earlier)
        assert not earlier.contradicts(later)

    def test_no_contradiction_date_and_time_period(self):
        later = Comparison(
            "the date $dentist became a licensed dentist was",
            sign=">",
            expression=date(2010, 1, 1),
        )
        earlier = Comparison(
            "the date $dentist became a licensed dentist was",
            sign="<",
            expression="2000 years",
        )
        assert not later.contradicts(earlier)
        assert not earlier.contradicts(later)

    def test_no_contradiction_irrelevant_quantities(self):
        more_cows = Comparison(
            "the number of cows $person owned was",
            sign=">",
            expression=10,
        )
        fewer_horses = Comparison(
            "the number of horses $person owned was",
            sign="<",
            expression=3,
        )
        assert not more_cows.contradicts(fewer_horses)
        assert not fewer_horses.contradicts(more_cows)

    def test_no_contradiction_of_predicate(self):
        more_cows = Comparison(
            "the number of cows $person owned was",
            sign=">",
            expression=10,
        )
        no_cows = Predicate("the number of cows $person owned was", truth=False)
        assert not more_cows.contradicts(no_cows)
        assert not no_cows.contradicts(more_cows)


class TestQuantities:
    def test_does_not_exclude_other_quantity(self):
        comparison = Comparison(
            "the distance between $place1 and $place2 was",
            sign=">",
            expression=Q_("20 miles"),
        )
        comparison_opposite = Comparison(
            "the distance between $place1 and $place2 was",
            sign="<",
            expression=Q_("30 miles"),
        )
        assert not comparison.contradicts(comparison_opposite)

    def test_convert_quantity_of_Comparison(self):
        comparison = Comparison(
            "the distance between $place1 and $place2 was",
            sign=">",
            expression=Q_("20 miles"),
        )
        comparison_km = Comparison(
            "the distance between $place1 and $place2 was",
            sign=">",
            expression=Q_("30 kilometers"),
        )
        assert comparison > comparison_km

    def test_quantity_comparison_to_predicate(self):
        distance = Comparison(
            "the distance between $place1 and $place2 was",
            sign=">",
            expression="20 miles",
        )
        predicate = Predicate("the distance between $place1 and $place2 was")
        assert not distance >= predicate



"""Create models of assertions accepted as factual by courts."""
from __future__ import annotations
from copy import deepcopy
from typing import ClassVar, Iterable, Iterator, List, Optional, Sequence, Tuple, Union

from anchorpoint.textselectors import TextQuoteSelector

from nettlesome.factors import Factor
from nettlesome.formatting import indented

from nettlesome.terms import (
    Comparable,
    ContextRegister,
    Explanation,
    TermSequence,
)
from nettlesome.predicates import Predicate
from nettlesome.statements import Statement


class Fact(Statement):
    r"""
    An assertion accepted as factual by a court.

    Often based on factfinding by a judge or jury.

    Facts may be "alleged" by a pleading, "supported" by evidence, or
    "found" to be factual by a jury verdict or a judge's finding of fact.

    :param predicate:
        a natural-language clause with zero or more slots
        to insert ``terms`` that are typically the
        subject and objects of the clause.

    :param terms:
        a series of :class:`Factor` objects that fill in
        the blank spaces in the ``predicate`` statement.

    :param name:
        an identifier for this object, often used if the object needs
        to be referred to multiple times in the process of composing
        other :class:`Factor` objects.

    :param standard_of_proof:
        a descriptor for the degree of certainty associated
        with the assertion in the ``predicate``.

    :param absent:
        whether the absence, rather than the presence, of the legal
        fact described above is being asserted.

    :param generic:
        whether this object could be replaced by another generic
        object of the same class without changing the truth of the
        :class:`Rule` in which it is mentioned.

    :attr standards_of_proof:
        a tuple with every allowable name for a standard of
        proof, in order from weakest to strongest.

        .. note:
            If any courts anywhere in a legal regime disagree about the
            relative strength of the various standards of proof, or if
            any court considers the order context-specific, then this
            approach of hard-coding their names and order will have to change.
    """

    standards_of_proof: ClassVar[Tuple[str, ...]] = (
        "scintilla of evidence",
        "substantial evidence",
        "preponderance of evidence",
        "clear and convincing",
        "beyond reasonable doubt",
    )

    def __init__(
        self,
        predicate: Predicate,
        terms: TermSequence = TermSequence(),
        name: Optional[str] = None,
        standard_of_proof: Optional[str] = None,
        absent: bool = False,
        generic: bool = False,
        anchors: Optional[List[TextQuoteSelector]] = None,
        truth: Optional[bool] = None,
    ):
        Statement.__init__(
            self,
            predicate=predicate,
            terms=terms,
            absent=absent,
            generic=generic,
            truth=truth,
        )
        self.standard_of_proof = standard_of_proof
        self.anchors = anchors or []
        if (
            self.standard_of_proof
            and self.standard_of_proof not in self.standards_of_proof
        ):
            raise ValueError(
                f"standard of proof must be one of {self.standards_of_proof} or None."
            )

    @property
    def wrapped_string(self):
        text = super().wrapped_string
        if self.standard_of_proof:
            text += "\n" + indented(f"by the STANDARD {self.standard_of_proof}")
        return text

    def __str__(self):
        """Create one-line string representation for inclusion in other Facts."""
        unwrapped = self.predicate._add_truth_to_content(self.content)
        standard = (
            f"by the standard {self.standard_of_proof}, "
            if self.standard_of_proof
            else ""
        )
        string = f"{standard}{unwrapped}"
        return Comparable.__str__(self).format(string).replace(",,", ",")

    @property
    def content(self) -> str:
        return str(self.predicate._content_with_terms(self.terms))

    def _means_if_concrete(
        self, other: Comparable, context: Explanation
    ) -> Iterator[Explanation]:
        if self.standard_of_proof == other.__dict__.get("standard_of_proof"):
            yield from super()._means_if_concrete(other, context)

    def __len__(self):
        return len(self.terms)

    def _implies_if_concrete(
        self, other: Comparable, context: ContextRegister
    ) -> Iterator[ContextRegister]:
        """
        Test if ``self`` implies ``other``, assuming they are not ``generic``.

        :returns:
            whether ``self`` implies ``other`` under the given assumption.
        """
        if (
            isinstance(other, self.__class__)
            and bool(self.standard_of_proof) == bool(other.standard_of_proof)
            and not (
                self.standard_of_proof
                and (
                    self.standards_of_proof.index(self.standard_of_proof)
                    < self.standards_of_proof.index(other.standard_of_proof)
                )
            )
        ):
            yield from super()._implies_if_concrete(other, context)

    def negated(self) -> Fact:
        """Return copy of self with opposite truth value."""
        result = deepcopy(self)
        result.predicate = result.predicate.negated()
        return result


def build_fact(
    predicate: Predicate,
    indices: Optional[Union[int, Sequence[int]]] = None,
    case_factors: Optional[Union[Factor, Sequence[Factor]]] = None,
    name: Optional[str] = None,
    standard_of_proof: Optional[str] = None,
    absent: bool = False,
    generic: bool = False,
):
    r"""
    Build a :class:`.Fact` with generics selected from a list.

    :param predicate:
        a natural-language clause with zero or more slots
        to insert ``terms`` that are typically the
        subject and objects of the clause.

    :param terms:
        a series of integer indices of generic factors to
        fill in the blanks in the :class:`.Predicate`

    :param name:
        an identifier for this object, often used if the object needs
        to be referred to multiple times in the process of composing
        other :class:`.Factor` objects

    :param standard_of_proof:
        a descriptor for the degree of certainty associated
        with the assertion in the :class:`.Predicate`

    :param absent:
        whether the absence, rather than the presence, of the legal
        fact described above is being asserted.

    :param generic:
        whether this object could be replaced by another generic
        object of the same class without changing the truth of the
        :class:`Rule` in which it is mentioned.

    :param case_factors:
        a series of :class:`.Factor`\s that have already been mentioned
        in the :class:`.Opinion`. They are available for composing the
        new :class:`.Factor` object and don't need to be recreated.
    """
    if not indices:
        indices = range(len(predicate))
    if isinstance(indices, int):
        indices = (indices,)

    case_factors = case_factors or ()
    if not isinstance(case_factors, Iterable):
        wrapped_factors = (case_factors,)
    else:
        wrapped_factors = tuple(case_factors)

    terms = TermSequence([wrapped_factors[i] for i in indices])
    return Fact(
        predicate=predicate,
        terms=terms,
        name=name,
        standard_of_proof=standard_of_proof,
        absent=absent,
        generic=generic,
    )



"""Fake download client to get fake Enactments for testing."""

from __future__ import annotations

import datetime
import json
from pathlib import Path
from typing import Dict, Optional, Union

from legislice.download import Client, normalize_path, LegislicePathError, RawEnactment
from legislice.enactments import CrossReference, Enactment


# A dict indexing responses by iso-format date strings.
ResponsesByDate = Dict[str, Dict]
ResponsesByDateByPath = Dict[str, Dict[str, Dict]]


class FakeClient(Client):
    """
    Repository for mocking API responses locally.

    Imitates the interface of :class:`legislice.download.Client`
    """

    def __init__(self, responses: ResponsesByDateByPath):
        self.responses = responses
        self.coverage: Dict[str, Dict[str, Union[datetime.date, str]]] = {
            "/us/const": {
                "latest_heading": "United States Constitution",
                "first_published": datetime.date(1788, 6, 21),
                "earliest_in_db": datetime.date(1788, 6, 21),
            },
            "/us/usc": {
                "latest_heading": "United States Code (USC)",
                "first_published": datetime.date(1926, 6, 30),
                "earliest_in_db": datetime.date(2013, 7, 18),
                "latest_in_db": datetime.date(2020, 8, 8),
            },
            "/test/acts": {
                "latest_heading": "Test Acts",
                "first_published": datetime.date(1935, 4, 1),
                "earliest_in_db": datetime.date(1935, 4, 1),
                "latest_in_db": datetime.date(2013, 7, 18),
            },
        }
        self.update_coverage_from_api = False

    @classmethod
    def from_file(cls, filename: str) -> FakeClient:
        parent_directory = Path(__file__).resolve().parents[2]
        responses_filepath = parent_directory / "example_data" / "responses" / filename
        with open(responses_filepath, "r") as f:
            responses = json.load(f)
        return cls(responses)

    def get_entry_closest_to_cited_path(self, path: str) -> Optional[ResponsesByDate]:
        path = normalize_path(path)
        if self.responses.get(path):
            return self.responses[path]
        branches_that_start_path = [
            entry for entry in self.responses.keys() if path.startswith(entry)
        ]
        if not branches_that_start_path:
            return None
        name_of_best_entry = max(branches_that_start_path, key=len)
        return self.responses[name_of_best_entry]

    def search_tree_for_path(
        self, path: str, branch: Dict
    ) -> Optional[ResponsesByDate]:
        path = normalize_path(path)
        if branch["node"] == path:
            return branch
        branches_that_start_path = [
            nested_node
            for nested_node in branch["children"]
            if path.startswith(nested_node["node"])
        ]
        if branches_that_start_path:
            return self.search_tree_for_path(
                path=path, branch=branches_that_start_path[0]
            )
        return None

    def fetch(self, query: str, date: Union[datetime.date, str] = "") -> RawEnactment:
        """
        Fetches data about legislation at specified path and date from Client's assigned API root.

        :param path:
            A path to the desired legislation section using the United States Legislation Markup
            tree-like citation format. Examples: /us/const/amendment/IV, /us/usc/t17/s103

        :param date:
            A date when the desired version of the legislation was in effect. This does not need to
            be the "effective date" or the first date when the version was in effect. However, if
            you select a date when two versions of the provision were in effect at the same time,
            you will be given the version that became effective later.

        :returns:
            A fake JSON response in the format of the Legislice API.
        """
        responses = self.get_entry_closest_to_cited_path(query)
        if not responses:
            raise LegislicePathError(f"No enacted text found for query {query}")

        if isinstance(date, datetime.date):
            date = date.isoformat()

        if not date:
            selected_date = max(responses.keys())
        else:
            versions_not_later_than_query = [
                version_date
                for version_date in responses.keys()
                if version_date <= date
            ]
            if not versions_not_later_than_query:
                raise ValueError(
                    f"No enacted text found for query {query} after date {date}"
                )
            selected_date = max(versions_not_later_than_query)

        selected_version = responses[selected_date]

        result = self.search_tree_for_path(path=query, branch=selected_version)
        if not result:
            raise LegislicePathError(
                f"No enacted text found for query {query} after date {date}"
            )
        return result

    def _fetch_from_url(self, url: str) -> None:
        raise RuntimeError("Network access not allowed from FakeClient")

    def read(
        self,
        query: Union[str, CrossReference],
        date: Union[datetime.date, str] = "",
    ) -> Enactment:
        """
        Use text expansion, unlike a real client.
        """
        raw_enactment = self.fetch(query=query, date=date)
        enactment = self.read_from_json(raw_enactment, use_text_expansion=True)
        enactment.select_all()
        return enactment
